# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import List, Sequence  # noqa: F401

from snapcraft.internal.errors import SnapcraftError as _SnapcraftError
from snapcraft.internal.errors import SnapcraftException


class RemoteBuildBaseError(_SnapcraftError):
    """Base Exception for all remote build related exceptions."""


class RemoteBuildNotFoundError(RemoteBuildBaseError):

    fmt = (
        "Remote build not found, please make sure project {name!r} and build "
        "number {req_number} are correct. If this is an old build, it may "
        "have expired."
    )

    def __init__(self, *, name: str, req_number: int) -> None:
        super().__init__(name=name, req_number=req_number)


class NoLaunchpadUsernameError(RemoteBuildBaseError):

    fmt = "Please set your Launchpad username using the '--user' option."


class NotGitRepositoryError(RemoteBuildBaseError):

    fmt = "Current directory is not a git repository."


class GitNotFoundVersionError(RemoteBuildBaseError):

    fmt = (
        "This remote-build project requires `git` to be installed "
        "because of it use of `version: git`.  Either install `git`, "
        "set the `version` statically, or use the `snapcraftctl set-version` "
        "part scriptlet with `adopt-info`."
    )


class GitNotFoundProviderError(RemoteBuildBaseError):

    fmt = (
        "The remote build provider ({provider!r}) requires "
        "the use of the `git` utility, and `git` is not installed. "
        "Please install `git` and run the command again."
    )

    def __init__(self, *, provider: str) -> None:
        super().__init__(provider=provider)


class BaseRequiredError(RemoteBuildBaseError):

    fmt = (
        "Remote build currently requires that the project uses bases.\n"
        "Please specify an appropriate `base` keyword and try again."
    )


class RemoteBuilderNotSupportedError(RemoteBuildBaseError):

    fmt = (
        "Remote builder {provider!r} is not supported, please choose "
        "a different one and try again."
    )

    def __init__(self, *, provider: str) -> None:
        super().__init__(provider=provider)


class RemoteBuildTimeoutError(RemoteBuildBaseError):

    fmt = "Remote build exceeded configured timeout."


class RemoteBuilderError(RemoteBuildBaseError):

    fmt = "Remote builder failed with error: {builder_error!r}"

    def __init__(self, *, builder_error: str) -> None:
        super().__init__(builder_error=builder_error)


class UnsupportedArchitectureError(RemoteBuildBaseError):

    fmt = (
        "The following architectures are not supported by the remote builder: "
        "{archs}.\nPlease remove them from the architecture list "
        "and try again."
    )

    def __init__(self, *, archs: List[str]) -> None:
        super().__init__(archs=", ".join(archs))


class UnsupportedVersionScriptError(RemoteBuildBaseError):

    fmt = (
        "Remote-build does not support the use of `version-script`.\n"
        "Please use `snapcraftctl set-version` part scriptlet with "
        "`adopt-info`, or set `version` statically."
    )


class AcceptPublicUploadError(RemoteBuildBaseError):

    fmt = (
        "Remote build needs explicit acknowledgement that data sent to build servers "
        "is public.\n"
        "In non-interactive runs, please use option --launchpad-accept-public-upload."
    )


class LaunchpadGitPushError(SnapcraftException):
    def __init__(self, *, command: str, exit_code: int) -> None:
        self._command = command
        self._exit_code = exit_code

    def get_brief(self) -> str:
        return "Failed to push sources to Launchpad."

    def get_details(self) -> str:
        return f"Command {self._command!r} failed with exit code {self._exit_code!r}."

    def get_resolution(self) -> str:
        return "Verify connectivity to https://git.launchpad.net and retry build."


class LaunchpadHttpsError(SnapcraftException):
    def get_brief(self) -> str:
        return "Failed to connect to Launchpad API service."

    def get_resolution(self) -> str:
        return "Verify connectivity to https://api.launchpad.net and retry build."



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Tim Süberkrüb
# Copyright (C) 2017-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
import subprocess

import fixtures
from testtools.matchers import Equals, MatchesRegex

from snapcraft.internal import sources
from tests import unit


def get_side_effect(original_call):
    def side_effect(cmd, *args, **kwargs):
        if cmd[0] == "7z":
            return "".encode()
        return original_call(cmd, *args, **kwargs)

    return side_effect


class Test7z(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.source_dir = "7z-source-dir"
        self.test_7z_file_path = "fake-7z-file.7z"

        os.makedirs(self.source_dir)
        open(self.test_7z_file_path, "w").close()

        self.fake_check_output = self.useFixture(
            fixtures.MockPatch(
                "subprocess.check_output",
                side_effect=get_side_effect(subprocess.check_output),
            )
        )

    def test_pull_7z_file_must_extract(self):
        seven_zip_source = sources.SevenZip(self.test_7z_file_path, self.source_dir)
        seven_zip_source.pull()

    def test_extract_and_keep_7zfile(self):
        seven_zip_source = sources.SevenZip(self.test_7z_file_path, self.source_dir)
        # This is the first step done by pull. We don't call pull to call the
        # second step with a different keep_7z value.
        shutil.copy2(seven_zip_source.source, seven_zip_source.source_dir)
        seven_zip_source.provision(dst=self.source_dir, keep_7z=True)

    def test_has_source_handler_entry(self):
        self.assertTrue(sources._source_handler["7z"] is sources.SevenZip)

    def test_pull_failure(self):
        self.fake_check_output.mock.side_effect = subprocess.CalledProcessError(1, [])

        os.makedirs("dst")
        seven_zip = sources.SevenZip(self.test_7z_file_path, "dst")
        raised = self.assertRaises(sources.errors.SnapcraftPullError, seven_zip.pull)
        self.assertThat(
            raised.command, MatchesRegex("7z x .*/{}".format(self.test_7z_file_path))
        )
        self.assertThat(raised.exit_code, Equals(1))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016, 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os

from pyramid import response

from tests.fake_servers import base

logger = logging.getLogger(__name__)


class FakeStoreUploadServer(base.BaseFakeServer):
    def configure(self, configurator):
        configurator.add_route(
            "unscanned-upload", "/unscanned-upload/", request_method="POST"
        )
        configurator.add_view(self.unscanned_upload, route_name="unscanned-upload")

    def unscanned_upload(self, request):
        logger.info("Handling upload request")
        if "UPDOWN_BROKEN" in os.environ:
            response_code = 500
            content_type = "text/plain"
            payload = b"Broken"
        else:
            response_code = 200
            content_type = "application/octet-stream"
            payload = json.dumps({"upload_id": "test-upload-id"}).encode()
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
from unittest import mock

from testtools.matchers import Equals, HasLength

from snapcraft.internal import errors
from snapcraft.plugins.v1 import meson
from tests import unit

from . import PluginsV1BaseTestCase


class MesonPluginPropertiesTest(unit.TestCase):
    def test_schema(self):
        """Test validity of the Meson Plugin schema"""
        schema = meson.MesonPlugin.schema()

        # Verify the presence of all properties
        properties = schema["properties"]
        self.assertTrue(
            "meson-parameters" in properties,
            'Expected "meson-parameters" to be included in ' "properties",
        )

        meson_parameters = properties["meson-parameters"]

        self.assertTrue(
            "type" in meson_parameters,
            'Expected "type" to be included in "meson-parameters"',
        )
        self.assertThat(
            meson_parameters["type"],
            Equals("array"),
            'Expected "meson-parameters" "type" to be "array", '
            'but it was "{}"'.format(meson_parameters["type"]),
        )

        self.assertTrue(
            "minitems" in meson_parameters,
            'Expected "minitems" to be included in "meson-parameters"',
        )
        self.assertThat(
            meson_parameters["minitems"],
            Equals(1),
            'Expected "meson-parameters" "minitems" to be 1, but '
            'it was "{}"'.format(meson_parameters["minitems"]),
        )

        self.assertTrue(
            "uniqueItems" in meson_parameters,
            'Expected "uniqueItems" to be included in "meson-parameters"',
        )
        self.assertTrue(
            meson_parameters["uniqueItems"],
            'Expected "meson-parameters" "uniqueItems" to be "True"',
        )

    def test_get_pull_properties(self):
        expected_pull_properties = ["meson-version"]
        resulting_pull_properties = meson.MesonPlugin.get_pull_properties()

        self.assertThat(
            resulting_pull_properties, HasLength(len(expected_pull_properties))
        )

        for property in expected_pull_properties:
            self.assertIn(property, resulting_pull_properties)

    def test_get_build_properties(self):
        expected_build_properties = ["meson-parameters"]
        resulting_build_properties = meson.MesonPlugin.get_build_properties()

        self.assertThat(
            resulting_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, resulting_build_properties)


class MesonPluginBaseTest(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        class Options:
            """Internal Options Class matching the Meson plugin"""

            meson_parameters = []
            meson_version = []

        self.options = Options()

    @mock.patch("subprocess.check_call")
    def test_pull(self, check_call_mock):
        """Test pulling with the default settings."""
        plugin = meson.MesonPlugin("test-part", self.options, self.project)

        plugin.pull()

        check_call_mock.assert_called_once_with(
            ["python3", "-m", "pip", "install", "-U", "meson"]
        )

    @mock.patch("subprocess.check_call")
    def test_pull_meson_version(self, check_call_mock):
        """Test pulling with the default settings."""
        self.options.meson_version = "1.0"
        plugin = meson.MesonPlugin("test-part", self.options, self.project)

        plugin.pull()

        check_call_mock.assert_called_once_with(
            ["python3", "-m", "pip", "install", "-U", "meson==1.0"]
        )

    @mock.patch(
        "subprocess.check_call",
        side_effect=subprocess.CalledProcessError(cmd=["meson-install"], returncode=1),
    )
    def test_pull_meson_installation_fails(self, check_call_mock):
        """Test pulling with the default settings."""
        self.options.meson_version = "1.0"
        plugin = meson.MesonPlugin("test-part", self.options, self.project)

        self.assertRaises(errors.SnapcraftPluginCommandError, plugin.pull)

    @mock.patch.object(meson.MesonPlugin, "run")
    def test_build(self, run_mock):
        """Test building via meson and check for known calls and destdir"""
        plugin = meson.MesonPlugin("test-part", self.options, self.project)

        plugin.build()

        env = os.environ.copy()
        env["DESTDIR"] = plugin.installdir

        self.assertThat(run_mock.call_count, Equals(3))
        run_mock.assert_has_calls(
            [
                mock.call(["meson", plugin.snapbuildname]),
                mock.call(["ninja"], cwd=plugin.mesonbuilddir),
                mock.call(["ninja", "install"], env=env, cwd=plugin.mesonbuilddir),
            ]
        )

    @mock.patch.object(meson.MesonPlugin, "run")
    def test_build_with_parameters(self, run_mock):
        """Test with parameters"""
        self.options.meson_parameters = ["--strip"]
        plugin = meson.MesonPlugin("test-part", self.options, self.project)

        plugin.build()

        env = os.environ.copy()
        env["DESTDIR"] = plugin.installdir

        self.assertThat(run_mock.call_count, Equals(3))
        run_mock.assert_has_calls(
            [
                mock.call(["meson", "--strip", plugin.snapbuildname]),
                mock.call(["ninja"], cwd=plugin.mesonbuilddir),
                mock.call(["ninja", "install"], env=env, cwd=plugin.mesonbuilddir),
            ]
        )


class MesonPluginUnsupportedBase(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        self.project._snap_meta.base = "unsupported-base"

        class Options:
            source = "dir"

        self.options = Options()

    def test_unsupported_base_raises(self):
        self.assertRaises(
            errors.PluginBaseError,
            meson.MesonPlugin,
            "test-part",
            self.options,
            self.project,
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from collections import OrderedDict
from unittest.mock import call, patch

import fixtures
from testtools.matchers import Contains, Equals

from snapcraft import extractors, plugins
from snapcraft.internal import elf, errors, states, steps
from tests import fixture_setup, unit


class StateBaseTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.get_pull_properties_mock = self.useFixture(
            fixtures.MockPatch(
                "snapcraft.plugins.v1.PluginV1.get_pull_properties", return_value=[]
            )
        ).mock

        self.get_build_properties_mock = self.useFixture(
            fixtures.MockPatch(
                "snapcraft.plugins.v1.PluginV1.get_build_properties", return_value=[]
            )
        ).mock

        self.handler = self.load_part("test_part")
        self.handler.makedirs()

        self.get_elf_files_mock = self.useFixture(
            fixtures.MockPatch(
                "snapcraft.internal.elf.get_elf_files", return_value=frozenset()
            )
        ).mock

        self.useFixture(
            fixtures.MockPatch(
                "snapcraft.internal.xattrs.read_origin_stage_package", return_value=None
            )
        )


class PullStateTestCase(StateBaseTestCase):
    def test_pull_build_packages_without_grammar_properties(self):
        self.handler = self.load_part(
            "test_part", part_properties={"build-packages": ["package1"]}
        )
        self.handler.mark_pull_done()
        state = self.handler.get_pull_state()

        self.assertTrue(type(state) is states.PullState)
        self.assertThat(state.assets.get("build-packages"), Equals({"package1"}))

    def test_pull_build_packages_with_grammar_properties(self):
        self.handler = self.load_part(
            "test_part",
            part_properties={
                "build-packages": [
                    {"on amd64": ["package1"]},
                    {"on i386": ["package2"]},
                    "package3",
                ],
                "build-snaps": [
                    {"on amd64": ["snap1"]},
                    {"on i386": ["snap2"]},
                    "snap3",
                ],
            },
        )
        self.handler.mark_pull_done()
        state = self.handler.get_pull_state()

        self.assertTrue(type(state) is states.PullState)
        self.assertThat(
            state.assets.get("build-packages"), Equals(set(["package1", "package3"]))
        )
        self.assertThat(
            state.assets.get("build-snaps"), Equals(set(["snap1", "snap3"]))
        )


class StateTestCase(StateBaseTestCase):
    @patch("snapcraft.internal.repo.Repo")
    def test_pull_state(self, repo_mock):
        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))
        repo_mock.get_installed_build_packages.return_value = []

        self.handler.pull()

        self.assertThat(self.handler.latest_step(), Equals(steps.PULL))
        self.assertThat(self.handler.next_step(), Equals(steps.BUILD))
        state = self.handler.get_pull_state()

        self.assertTrue(state, "Expected pull to save state YAML")
        self.assertTrue(type(state) is states.PullState)
        self.assertTrue(type(state.properties) is OrderedDict)
        self.assertThat(len(state.properties), Equals(11))
        for expected in [
            "source",
            "source-branch",
            "source-commit",
            "source-depth",
            "source-subdir",
            "source-tag",
            "source-type",
            "plugin",
            "stage-packages",
            "parse-info",
            "override-pull",
        ]:
            self.assertTrue(expected in state.properties)
        self.assertTrue(type(state.project_options) is OrderedDict)
        self.assertTrue("deb_arch" in state.project_options)

    @patch("snapcraft.internal.repo.Repo")
    def test_pull_state_with_extracted_metadata(self, repo_mock):
        self.handler = self.load_part(
            "test_part",
            part_properties={"source": ".", "parse-info": ["metadata-file"]},
        )

        # Create metadata file
        open("metadata-file", "w").close()

        def _fake_extractor(file_path, workdir):
            return extractors.ExtractedMetadata(
                common_id="test_common_id",
                summary="test summary",
                description="test description",
                icon="/test/path",
                desktop_file_paths=[
                    "usr/share/applications/com.example.test/app.desktop"
                ],
            )

        self.useFixture(fixture_setup.FakeMetadataExtractor("fake", _fake_extractor))

        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))
        repo_mock.get_installed_build_packages.return_value = []

        self.handler.pull()

        self.assertThat(self.handler.latest_step(), Equals(steps.PULL))
        self.assertThat(self.handler.next_step(), Equals(steps.BUILD))
        state = self.handler.get_pull_state()

        self.assertTrue(state, "Expected pull to save state YAML")
        self.assertTrue(type(state) is states.PullState)
        self.assertTrue(type(state.properties) is OrderedDict)
        self.assertThat(len(state.properties), Equals(11))
        for expected in [
            "source",
            "source-branch",
            "source-commit",
            "source-depth",
            "source-subdir",
            "source-tag",
            "source-type",
            "plugin",
            "stage-packages",
            "parse-info",
            "override-pull",
        ]:
            self.assertThat(state.properties, Contains(expected))
        self.assertTrue(type(state.project_options) is OrderedDict)
        self.assertThat(state.project_options, Contains("deb_arch"))

        self.assertTrue(type(state.extracted_metadata) is OrderedDict)
        for expected in ("metadata", "files"):
            self.assertThat(state.extracted_metadata, Contains(expected))
        metadata = state.extracted_metadata["metadata"]
        self.assertThat(metadata.get_common_id(), Equals("test_common_id"))
        self.assertThat(metadata.get_summary(), Equals("test summary"))
        self.assertThat(metadata.get_description(), Equals("test description"))
        self.assertThat(metadata.get_icon(), Equals("/test/path"))
        self.assertThat(
            metadata.get_desktop_file_paths(),
            Equals(["usr/share/applications/com.example.test/app.desktop"]),
        )
        files = state.extracted_metadata["files"]
        self.assertThat(
            files, Equals([os.path.join(self.handler.part_source_dir, "metadata-file")])
        )

    @patch("snapcraft.internal.repo.Repo")
    def test_pull_state_with_scriptlet_metadata(self, repo_mock):
        self.handler = self.load_part(
            "test_part",
            part_properties={
                "override-pull": "snapcraftctl set-version override-version"
            },
        )

        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))
        repo_mock.get_installed_build_packages.return_value = []

        self.handler.pull()

        self.assertThat(self.handler.latest_step(), Equals(steps.PULL))
        self.assertThat(self.handler.next_step(), Equals(steps.BUILD))
        state = self.handler.get_pull_state()

        self.assertTrue(state, "Expected pull to save state YAML")
        self.assertTrue(type(state) is states.PullState)
        self.assertTrue(type(state.properties) is OrderedDict)
        self.assertThat(len(state.properties), Equals(11))
        for expected in [
            "source",
            "source-branch",
            "source-commit",
            "source-depth",
            "source-subdir",
            "source-tag",
            "source-type",
            "plugin",
            "stage-packages",
            "parse-info",
            "override-pull",
        ]:
            self.assertThat(state.properties, Contains(expected))
        self.assertThat(
            state.properties["override-pull"],
            Equals("snapcraftctl set-version override-version"),
        )

        metadata = state.scriptlet_metadata
        self.assertThat(metadata.get_version(), Equals("override-version"))

    def test_pull_state_with_properties(self):
        self.get_pull_properties_mock.return_value = ["foo"]
        self.handler.plugin.options.foo = "bar"
        self.handler._part_properties["foo"] = "bar"

        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        self.handler.pull()

        self.assertThat(self.handler.latest_step(), Equals(steps.PULL))
        self.assertThat(self.handler.next_step(), Equals(steps.BUILD))
        state = self.handler.get_pull_state()

        self.assertTrue(state, "Expected pull to save state YAML")
        self.assertTrue(type(state) is states.PullState)
        self.assertTrue(type(state.properties) is OrderedDict)
        if isinstance(self.handler.plugin, plugins.v1.PluginV1):
            self.assertTrue("foo" in state.properties)
            self.assertThat(state.properties["foo"], Equals("bar"))
        self.assertTrue(type(state.project_options) is OrderedDict)
        self.assertTrue("deb_arch" in state.project_options)

    @patch.object(plugins.v1.nil.NilPlugin, "clean_pull")
    def test_clean_pull_state(self, mock_clean_pull):
        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        self.handler.pull()

        self.handler.clean_pull()

        # Verify that the plugin had clean_pull() called
        if isinstance(self.handler.plugin, plugins.v1.PluginV1):
            mock_clean_pull.assert_called_once_with()

        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

    def test_build_state(self):
        if not isinstance(self.handler.plugin, plugins.v1.PluginV1):
            self.skipTest("Test designed for PluginV1")

        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        self.handler.build()

        self.assertThat(self.handler.latest_step(), Equals(steps.BUILD))
        self.assertThat(self.handler.next_step(), Equals(steps.STAGE))
        state = self.handler.get_build_state()

        self.assertTrue(state, "Expected build to save state YAML")
        self.assertTrue(type(state) is states.BuildState)
        self.assertTrue(type(state.properties) is OrderedDict)
        self.assertThat(len(state.properties), Equals(6))
        for expected in [
            "after",
            "build-attributes",
            "build-packages",
            "disable-parallel",
            "organize",
            "override-build",
        ]:
            self.assertTrue(expected in state.properties)
        self.assertTrue(type(state.project_options) is OrderedDict)
        self.assertTrue("deb_arch" in state.project_options)

    def test_build_state_with_extracted_metadata(self):
        self.handler = self.load_part(
            "test_part", part_properties={"parse-info": ["metadata-file"]}
        )

        # Create metadata file
        open(os.path.join(self.handler.part_source_dir, "metadata-file"), "w").close()

        def _fake_extractor(file_path, workdir):
            return extractors.ExtractedMetadata(
                common_id="test_common_id",
                summary="test summary",
                description="test description",
                icon="/test/path",
                desktop_file_paths=[
                    "usr/share/applications/com.example.test/app.desktop"
                ],
            )

        self.useFixture(fixture_setup.FakeMetadataExtractor("fake", _fake_extractor))

        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        self.handler.build()

        self.assertThat(self.handler.latest_step(), Equals(steps.BUILD))
        self.assertThat(self.handler.next_step(), Equals(steps.STAGE))
        state = self.handler.get_build_state()

        self.assertTrue(state, "Expected build to save state YAML")
        self.assertTrue(type(state) is states.BuildState)
        self.assertTrue(type(state.properties) is OrderedDict)
        self.assertThat(len(state.properties), Equals(6))
        for expected in [
            "after",
            "build-attributes",
            "build-packages",
            "disable-parallel",
            "organize",
            "override-build",
        ]:
            self.assertThat(state.properties, Contains(expected))
        self.assertTrue(type(state.project_options) is OrderedDict)
        self.assertThat(state.project_options, Contains("deb_arch"))

        self.assertTrue(type(state.extracted_metadata) is OrderedDict)
        for expected in ("metadata", "files"):
            self.assertThat(state.extracted_metadata, Contains(expected))
        metadata = state.extracted_metadata["metadata"]
        self.assertThat(metadata.get_common_id(), Equals("test_common_id"))
        self.assertThat(metadata.get_summary(), Equals("test summary"))
        self.assertThat(metadata.get_description(), Equals("test description"))
        self.assertThat(metadata.get_icon(), Equals("/test/path"))
        self.assertThat(
            metadata.get_desktop_file_paths(),
            Equals(["usr/share/applications/com.example.test/app.desktop"]),
        )
        files = state.extracted_metadata["files"]
        self.assertThat(
            files, Equals([os.path.join(self.handler.part_build_dir, "metadata-file")])
        )

    def test_build_state_with_scriptlet_metadata(self):
        self.handler = self.load_part(
            "test_part",
            part_properties={
                "override-build": "snapcraftctl set-version override-version"
            },
        )

        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        self.handler.pull()
        self.handler.build()

        self.assertThat(self.handler.latest_step(), Equals(steps.BUILD))
        self.assertThat(self.handler.next_step(), Equals(steps.STAGE))
        state = self.handler.get_build_state()

        self.assertTrue(state, "Expected build to save state YAML")
        self.assertTrue(type(state) is states.BuildState)
        self.assertTrue(type(state.properties) is OrderedDict)
        self.assertThat(len(state.properties), Equals(6))
        for expected in [
            "after",
            "build-attributes",
            "build-packages",
            "disable-parallel",
            "organize",
            "override-build",
        ]:
            self.assertThat(state.properties, Contains(expected))
        self.assertThat(
            state.properties["override-build"],
            Equals("snapcraftctl set-version override-version"),
        )

        metadata = state.scriptlet_metadata
        self.assertThat(metadata.get_version(), Equals("override-version"))

    def test_build_state_with_properties(self):
        self.get_build_properties_mock.return_value = ["foo"]
        self.handler.plugin.options.foo = "bar"
        self.handler._part_properties = {"foo": "bar"}

        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        self.handler.build()

        self.assertThat(self.handler.latest_step(), Equals(steps.BUILD))
        self.assertThat(self.handler.next_step(), Equals(steps.STAGE))
        state = self.handler.get_build_state()

        self.assertTrue(state, "Expected build to save state YAML")
        self.assertTrue(type(state) is states.BuildState)
        self.assertTrue(type(state.properties) is OrderedDict)
        self.assertTrue("foo" in state.properties)
        self.assertThat(state.properties["foo"], Equals("bar"))
        self.assertTrue(type(state.project_options) is OrderedDict)
        self.assertTrue("deb_arch" in state.project_options)

    @patch.object(plugins.v1.nil.NilPlugin, "clean_build")
    def test_clean_build_state(self, mock_clean_build):
        if not isinstance(self.handler.plugin, plugins.v1.PluginV1):
            self.skipTest("Test designed for PluginV1")

        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)

        self.handler.mark_done(steps.PULL)
        self.handler.build()

        self.handler.clean_build()

        # Verify that the plugin had clean_build() called for v1.
        if isinstance(self.handler.plugin, plugins.v1.PluginV1):
            mock_clean_build.assert_called_once_with()

        self.assertThat(self.handler.latest_step(), Equals(steps.PULL))
        self.assertThat(self.handler.next_step(), Equals(steps.BUILD))

    def test_stage_state(self):
        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        bindir = os.path.join(self.handler.part_install_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "1"), "w").close()
        open(os.path.join(bindir, "2"), "w").close()

        self.handler.mark_done(steps.BUILD)
        self.handler.stage()

        self.assertThat(self.handler.latest_step(), Equals(steps.STAGE))
        self.assertThat(self.handler.next_step(), Equals(steps.PRIME))
        state = self.handler.get_stage_state()

        self.assertTrue(state, "Expected stage to save state YAML")
        self.assertTrue(type(state) is states.StageState)
        self.assertTrue(type(state.files) is set)
        self.assertTrue(type(state.directories) is set)
        self.assertTrue(type(state.properties) is OrderedDict)
        self.assertThat(len(state.files), Equals(2))
        self.assertTrue("bin/1" in state.files)
        self.assertTrue("bin/2" in state.files)
        self.assertThat(len(state.directories), Equals(1))
        self.assertTrue("bin" in state.directories)
        self.assertTrue("stage" in state.properties)
        self.assertThat(state.properties["stage"], Equals(["*"]))
        self.assertTrue("filesets" in state.properties)
        self.assertThat(state.properties["filesets"], Equals({}))
        self.assertTrue(type(state.project_options) is OrderedDict)
        self.assertThat(len(state.project_options), Equals(0))

    def test_stage_state_with_scriptlet_metadata(self):
        self.handler = self.load_part(
            "test_part",
            part_properties={
                "override-stage": "snapcraftctl set-version override-version"
            },
        )

        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        self.handler.pull()
        self.handler.build()
        self.handler.stage()

        self.assertThat(self.handler.latest_step(), Equals(steps.STAGE))
        self.assertThat(self.handler.next_step(), Equals(steps.PRIME))
        state = self.handler.get_stage_state()

        self.assertTrue(state, "Expected stage to save state YAML")
        self.assertTrue(type(state) is states.StageState)
        self.assertTrue(type(state.files) is set)
        self.assertTrue(type(state.directories) is set)
        self.assertTrue(type(state.properties) is OrderedDict)
        self.assertThat(len(state.properties), Equals(3))
        for expected in ["stage", "filesets", "override-stage"]:
            self.assertThat(state.properties, Contains(expected))
        self.assertThat(
            state.properties["override-stage"],
            Equals("snapcraftctl set-version override-version"),
        )

        metadata = state.scriptlet_metadata
        self.assertThat(metadata.get_version(), Equals("override-version"))

    def test_stage_state_with_stage_keyword(self):
        self.handler.plugin.options.stage = ["bin/1"]
        self.handler._part_properties["stage"] = ["bin/1"]

        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        bindir = os.path.join(self.handler.part_install_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "1"), "w").close()
        open(os.path.join(bindir, "2"), "w").close()

        self.handler.mark_done(steps.BUILD)
        self.handler.stage()

        self.assertThat(self.handler.latest_step(), Equals(steps.STAGE))
        self.assertThat(self.handler.next_step(), Equals(steps.PRIME))
        state = self.handler.get_stage_state()

        self.assertTrue(state, "Expected stage to save state YAML")
        self.assertTrue(type(state) is states.StageState)
        self.assertTrue(type(state.files) is set)
        self.assertTrue(type(state.directories) is set)
        self.assertTrue(type(state.properties) is OrderedDict)
        self.assertThat(len(state.files), Equals(1))
        self.assertTrue("bin/1" in state.files)
        self.assertThat(len(state.directories), Equals(1))
        self.assertTrue("bin" in state.directories)
        self.assertTrue("stage" in state.properties)
        self.assertThat(state.properties["stage"], Equals(["bin/1"]))
        self.assertTrue(type(state.project_options) is OrderedDict)
        self.assertThat(len(state.project_options), Equals(0))

        self.assertThat(self.handler.latest_step(), Equals(steps.STAGE))
        self.assertThat(self.handler.next_step(), Equals(steps.PRIME))

    def test_clean_stage_state(self):
        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))
        bindir = os.path.join(self.stage_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "1"), "w").close()
        open(os.path.join(bindir, "2"), "w").close()

        self.handler.mark_done(steps.BUILD)

        self.handler.mark_done(
            steps.STAGE, states.StageState({"bin/1", "bin/2"}, {"bin"})
        )

        self.handler.clean_stage({})

        self.assertThat(self.handler.latest_step(), Equals(steps.BUILD))
        self.assertThat(self.handler.next_step(), Equals(steps.STAGE))
        self.assertFalse(os.path.exists(bindir))

    def test_clean_stage_state_multiple_parts(self):
        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        bindir = os.path.join(self.stage_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "1"), "w").close()
        open(os.path.join(bindir, "2"), "w").close()
        open(os.path.join(bindir, "3"), "w").close()

        self.handler.mark_done(steps.BUILD)

        self.handler.mark_done(
            steps.STAGE, states.StageState({"bin/1", "bin/2"}, {"bin"})
        )

        self.handler.clean_stage({})

        self.assertThat(self.handler.latest_step(), Equals(steps.BUILD))
        self.assertThat(self.handler.next_step(), Equals(steps.STAGE))
        self.assertFalse(os.path.exists(os.path.join(bindir, "1")))
        self.assertFalse(os.path.exists(os.path.join(bindir, "2")))
        self.assertTrue(
            os.path.exists(os.path.join(bindir, "3")),
            "Expected 'bin/3' to remain as it wasn't staged by this part",
        )

    def test_clean_stage_state_common_files(self):
        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        bindir = os.path.join(self.stage_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "1"), "w").close()
        open(os.path.join(bindir, "2"), "w").close()

        self.handler.mark_done(steps.BUILD)

        self.handler.mark_done(
            steps.STAGE, states.StageState({"bin/1", "bin/2"}, {"bin"})
        )

        self.handler.clean_stage({"other_part": states.StageState({"bin/2"}, {"bin"})})

        self.assertThat(self.handler.latest_step(), Equals(steps.BUILD))
        self.assertThat(self.handler.next_step(), Equals(steps.STAGE))
        self.assertFalse(os.path.exists(os.path.join(bindir, "1")))
        self.assertTrue(
            os.path.exists(os.path.join(bindir, "2")),
            "Expected 'bin/2' to remain as it's required by other parts",
        )

    def test_clean_stage_old_state(self):
        self.handler.mark_done(steps.STAGE, None)
        raised = self.assertRaises(
            errors.MissingStateCleanError, self.handler.clean_stage, {}
        )

        self.assertThat(raised.step, Equals(steps.STAGE))

    @patch("shutil.copy")
    def test_prime_state(self, mock_copy):
        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        bindir = os.path.join(self.handler.part_install_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "1"), "w").close()
        open(os.path.join(bindir, "2"), "w").close()

        self.handler.mark_done(steps.BUILD)
        self.handler.stage()
        self.handler.prime()

        self.assertThat(self.handler.latest_step(), Equals(steps.PRIME))
        self.assertRaises(errors.NoNextStepError, self.handler.next_step)
        self.get_elf_files_mock.assert_called_once_with(
            self.handler._project.prime_dir, {"bin/1", "bin/2"}
        )
        self.assertFalse(mock_copy.called)

        state = self.handler.get_prime_state()

        self.assertTrue(type(state) is states.PrimeState)
        self.assertTrue(type(state.files) is set)
        self.assertTrue(type(state.directories) is set)
        self.assertTrue(type(state.dependency_paths) is set)
        self.assertTrue(type(state.properties) is OrderedDict)
        self.assertThat(len(state.files), Equals(2))
        self.assertTrue("bin/1" in state.files)
        self.assertTrue("bin/2" in state.files)
        self.assertThat(len(state.directories), Equals(1))
        self.assertTrue("bin" in state.directories)
        self.assertThat(len(state.dependency_paths), Equals(0))
        self.assertTrue("prime" in state.properties)
        self.assertThat(state.properties["prime"], Equals(["*"]))
        self.assertTrue(type(state.project_options) is OrderedDict)
        self.assertThat(len(state.project_options), Equals(0))

    def test_prime_state_with_scriptlet_metadata(self):
        self.handler = self.load_part(
            "test_part",
            part_properties={
                "override-prime": "snapcraftctl set-version override-version"
            },
        )

        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        self.handler.pull()
        self.handler.build()
        self.handler.stage()
        self.handler.prime()

        self.assertThat(self.handler.latest_step(), Equals(steps.PRIME))
        self.assertRaises(errors.NoNextStepError, self.handler.next_step)
        state = self.handler.get_prime_state()

        self.assertTrue(state, "Expected prime to save state YAML")
        self.assertTrue(type(state) is states.PrimeState)
        self.assertTrue(type(state.files) is set)
        self.assertTrue(type(state.directories) is set)
        self.assertTrue(type(state.dependency_paths) is set)
        self.assertTrue(type(state.properties) is OrderedDict)
        self.assertThat(len(state.properties), Equals(2))
        for expected in ["prime", "override-prime"]:
            self.assertThat(state.properties, Contains(expected))
        self.assertThat(
            state.properties["override-prime"],
            Equals("snapcraftctl set-version override-version"),
        )

        metadata = state.scriptlet_metadata
        self.assertThat(metadata.get_version(), Equals("override-version"))

    @patch("shutil.copy")
    def test_prime_state_with_stuff_already_primed(self, mock_copy):
        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        bindir = os.path.join(self.handler.part_install_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "1"), "w").close()
        bindir = os.path.join(self.handler._project.prime_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "2"), "w").close()

        self.handler.mark_done(steps.BUILD)
        self.handler.stage()
        self.handler.prime()

        self.assertThat(self.handler.latest_step(), Equals(steps.PRIME))
        self.assertRaises(errors.NoNextStepError, self.handler.next_step)
        # bin/2 shouldn't be in this list as it was already primed by another
        # part.
        self.get_elf_files_mock.assert_called_once_with(
            self.handler._project.prime_dir, {"bin/1"}
        )
        self.assertFalse(mock_copy.called)

        state = self.handler.get_prime_state()

        self.assertTrue(type(state) is states.PrimeState)
        self.assertTrue(type(state.files) is set)
        self.assertTrue(type(state.directories) is set)
        self.assertTrue(type(state.dependency_paths) is set)
        self.assertTrue(type(state.properties) is OrderedDict)
        self.assertThat(len(state.files), Equals(1))
        self.assertTrue("bin/1" in state.files)
        self.assertThat(len(state.directories), Equals(1))
        self.assertTrue("bin" in state.directories)
        self.assertThat(len(state.dependency_paths), Equals(0))
        self.assertTrue("prime" in state.properties)
        self.assertThat(state.properties["prime"], Equals(["*"]))
        self.assertTrue(type(state.project_options) is OrderedDict)
        self.assertThat(len(state.project_options), Equals(0))

    @patch("snapcraft.internal.elf.ElfFile._extract_attributes")
    @patch("snapcraft.internal.elf.ElfFile.load_dependencies")
    @patch("snapcraft.internal.pluginhandler._migrate_files")
    def test_prime_state_with_dependencies(
        self, mock_migrate_files, mock_load_dependencies, mock_get_symbols
    ):
        mock_load_dependencies.return_value = {
            "/foo/bar/baz",
            "{}/lib1/installed".format(self.handler.part_install_dir),
            "{}/lib2/staged".format(self.handler._project.stage_dir),
            "{}/lib3/primed".format(self.handler._project.prime_dir),
        }
        self.get_elf_files_mock.return_value = frozenset(
            [
                elf.ElfFile(
                    path=os.path.join(self.handler._project.prime_dir, "bin", "1")
                ),
                elf.ElfFile(
                    path=os.path.join(self.handler._project.prime_dir, "bin", "2")
                ),
            ]
        )
        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        bindir = os.path.join(self.handler.part_install_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "1"), "w").close()
        open(os.path.join(bindir, "2"), "w").close()

        self.handler.mark_done(steps.BUILD)
        self.handler.stage()

        # Resetting for test clarity
        mock_migrate_files.reset_mock()

        self.handler.prime()

        self.assertThat(self.handler.latest_step(), Equals(steps.PRIME))
        self.assertRaises(errors.NoNextStepError, self.handler.next_step)
        self.get_elf_files_mock.assert_called_once_with(
            self.handler._project.prime_dir, {"bin/1", "bin/2"}
        )
        mock_migrate_files.assert_has_calls(
            [
                call(
                    {"bin/1", "bin/2"},
                    {"bin"},
                    self.handler._project.stage_dir,
                    self.handler._project.prime_dir,
                )
            ]
        )

        state = self.handler.get_prime_state()

        self.assertTrue(type(state) is states.PrimeState)
        self.assertTrue(type(state.files) is set)
        self.assertTrue(type(state.directories) is set)
        self.assertTrue(type(state.properties) is OrderedDict)
        self.assertThat(len(state.files), Equals(2))
        self.assertThat(state.dependency_paths, Equals({"lib3"}))
        self.assertTrue("bin/1" in state.files)
        self.assertTrue("bin/2" in state.files)
        self.assertThat(len(state.directories), Equals(1))
        self.assertTrue("bin" in state.directories)
        self.assertTrue("prime" in state.properties)
        self.assertThat(state.properties["prime"], Equals(["*"]))
        self.assertTrue(type(state.project_options) is OrderedDict)
        self.assertThat(len(state.project_options), Equals(0))

    @patch("snapcraft.internal.elf.ElfFile._extract_attributes")
    @patch("snapcraft.internal.elf.ElfFile.load_dependencies")
    @patch("snapcraft.internal.pluginhandler._migrate_files")
    def test_prime_state_missing_libraries(
        self, mock_migrate_files, mock_load_dependencies, mock_get_symbols
    ):
        self.handler = self.load_part("test_part")

        self.get_elf_files_mock.return_value = frozenset(
            [
                elf.ElfFile(
                    path=os.path.join(self.handler._project.prime_dir, "bin", "file")
                )
            ]
        )
        # Pretend we found a system dependency, as well as a part and stage
        # dependency.
        mock_load_dependencies.return_value = set(
            [
                "/foo/bar/baz",
                "{}/lib1/installed".format(self.handler.part_install_dir),
                "{}/lib2/staged".format(self.handler._project.stage_dir),
                "{}/lib3/primed".format(self.handler._project.prime_dir),
            ]
        )

        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        bindir = os.path.join(self.handler.part_install_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "file"), "w").close()

        self.handler.mark_done(steps.BUILD)
        self.handler.stage()
        mock_migrate_files.reset_mock()
        self.handler.prime()

        self.assertThat(self.handler.latest_step(), Equals(steps.PRIME))
        self.assertRaises(errors.NoNextStepError, self.handler.next_step)
        self.get_elf_files_mock.assert_called_once_with(
            self.handler._project.prime_dir, {"bin/file"}
        )
        # Verify that only the part's files were migrated-- not the system
        # dependency.
        mock_migrate_files.assert_called_once_with(
            {"bin/file"},
            {"bin"},
            self.handler._project.stage_dir,
            self.handler._project.prime_dir,
        )

        state = self.handler.get_prime_state()

        # Verify that only the primed paths were captured.
        # The rest should be considered missing.
        self.assertThat(state.dependency_paths, Equals({"lib3"}))

    @patch("snapcraft.internal.elf.ElfFile._extract_attributes")
    @patch("snapcraft.internal.elf.ElfFile.load_dependencies")
    @patch("snapcraft.internal.pluginhandler._migrate_files")
    def test_prime_state_with_shadowed_dependencies(
        self, mock_migrate_files, mock_load_dependencies, mock_get_symbols
    ):
        self.get_elf_files_mock.return_value = frozenset([elf.ElfFile(path="bin/1")])
        mock_load_dependencies.return_value = {
            f"{self.handler._project.prime_dir}/foo/bar/baz"
        }

        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        bindir = os.path.join(self.handler.part_install_dir, "bin")
        foobardir = os.path.join(self.handler.part_install_dir, "foo", "bar")
        os.makedirs(bindir)
        os.makedirs(foobardir)

        # Make a "binary" as well as a "library" at the same path as the one on
        # the system
        open(os.path.join(bindir, "1"), "w").close()
        open(os.path.join(foobardir, "baz"), "w").close()

        self.handler.mark_done(steps.BUILD)
        self.handler.stage()
        mock_migrate_files.reset_mock()
        self.handler.prime()

        self.assertThat(self.handler.latest_step(), Equals(steps.PRIME))
        self.assertRaises(errors.NoNextStepError, self.handler.next_step)
        self.get_elf_files_mock.assert_called_once_with(
            self.handler._project.prime_dir, {"bin/1", "foo/bar/baz"}
        )
        mock_migrate_files.assert_called_once_with(
            {"bin/1", "foo/bar/baz"},
            {"bin", "foo", "foo/bar"},
            self.handler._project.stage_dir,
            self.handler._project.prime_dir,
        )

        state = self.handler.get_prime_state()

        self.assertTrue(type(state) is states.PrimeState)
        self.assertThat(state.dependency_paths, Equals({"foo/bar"}))

    @patch("shutil.copy")
    def test_prime_state_with_prime_keyword(self, mock_copy):
        self.handler = self.load_part("test_part", part_properties={"prime": ["bin/1"]})

        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        bindir = os.path.join(self.handler.part_install_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "1"), "w").close()
        open(os.path.join(bindir, "2"), "w").close()

        self.handler.mark_done(steps.BUILD)
        self.handler.stage()
        self.handler.prime()

        self.assertThat(self.handler.latest_step(), Equals(steps.PRIME))
        self.assertRaises(errors.NoNextStepError, self.handler.next_step)
        self.get_elf_files_mock.assert_called_once_with(
            self.handler._project.prime_dir, {"bin/1"}
        )
        self.assertFalse(mock_copy.called)

        state = self.handler.get_prime_state()

        self.assertTrue(type(state) is states.PrimeState)
        self.assertTrue(type(state.files) is set)
        self.assertTrue(type(state.directories) is set)
        self.assertTrue(type(state.dependency_paths) is set)
        self.assertTrue(type(state.properties) is OrderedDict)
        self.assertThat(len(state.files), Equals(1))
        self.assertTrue("bin/1" in state.files)
        self.assertThat(len(state.directories), Equals(1))
        self.assertTrue("bin" in state.directories)
        self.assertThat(len(state.dependency_paths), Equals(0))
        self.assertTrue("prime" in state.properties)
        self.assertThat(state.properties["prime"], Equals(["bin/1"]))
        self.assertTrue(type(state.project_options) is OrderedDict)
        self.assertThat(len(state.project_options), Equals(0))

    def test_clean_prime_state(self):
        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))
        bindir = os.path.join(self.prime_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "1"), "w").close()
        open(os.path.join(bindir, "2"), "w").close()

        self.handler.mark_done(steps.STAGE)

        self.handler.mark_done(
            steps.PRIME, states.PrimeState({"bin/1", "bin/2"}, {"bin"})
        )

        self.handler.clean_prime({})

        self.assertThat(self.handler.latest_step(), Equals(steps.STAGE))
        self.assertThat(self.handler.next_step(), Equals(steps.PRIME))
        self.assertFalse(os.path.exists(bindir))

    def test_clean_prime_state_inconsistent_files(self):
        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        self.handler.mark_done(steps.STAGE)

        self.handler.mark_done(
            steps.PRIME, states.PrimeState({"bin/1", "bin/2"}, {"bin", "not-there"})
        )

        # Only create a subset of the files and directories that were marked as primed
        bindir = os.path.join(self.prime_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "1"), "w").close()

        self.handler.clean_prime({})

        self.assertThat(self.handler.latest_step(), Equals(steps.STAGE))
        self.assertThat(self.handler.next_step(), Equals(steps.PRIME))
        self.assertFalse(os.path.exists(bindir))

    def test_clean_prime_state_multiple_parts(self):
        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))
        bindir = os.path.join(self.prime_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "1"), "w").close()
        open(os.path.join(bindir, "2"), "w").close()
        open(os.path.join(bindir, "3"), "w").close()

        self.handler.mark_done(steps.STAGE)

        self.handler.mark_done(
            steps.PRIME, states.PrimeState({"bin/1", "bin/2"}, {"bin"})
        )

        self.handler.clean_prime({})

        self.assertThat(self.handler.latest_step(), Equals(steps.STAGE))
        self.assertThat(self.handler.next_step(), Equals(steps.PRIME))
        self.assertFalse(os.path.exists(os.path.join(bindir, "1")))
        self.assertFalse(os.path.exists(os.path.join(bindir, "2")))
        self.assertTrue(
            os.path.exists(os.path.join(bindir, "3")),
            "Expected 'bin/3' to remain as it wasn't primed by this part",
        )

    def test_clean_prime_state_common_files(self):
        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))
        bindir = os.path.join(self.prime_dir, "bin")
        os.makedirs(bindir)
        open(os.path.join(bindir, "1"), "w").close()
        open(os.path.join(bindir, "2"), "w").close()

        self.handler.mark_done(steps.STAGE)

        self.handler.mark_done(
            steps.PRIME, states.PrimeState({"bin/1", "bin/2"}, {"bin"})
        )

        self.handler.clean_prime({"other_part": states.PrimeState({"bin/2"}, {"bin"})})

        self.assertThat(self.handler.latest_step(), Equals(steps.STAGE))
        self.assertThat(self.handler.next_step(), Equals(steps.PRIME))
        self.assertFalse(os.path.exists(os.path.join(bindir, "1")))
        self.assertTrue(
            os.path.exists(os.path.join(bindir, "2")),
            "Expected 'bin/2' to remain as it's required by other parts",
        )

    def test_clean_prime_old_state(self):
        self.handler.mark_done(steps.PRIME, None)
        raised = self.assertRaises(
            errors.MissingStateCleanError, self.handler.clean_prime, {}
        )

        self.assertThat(raised.step, Equals(steps.PRIME))


class TestStateFileMigration:
    scenarios = [(step.name, dict(step=step)) for step in steps.STEPS]

    def test_state_file_migration(self, tmp_work_path, step):
        part_name = "foo"

        part_dir = tmp_work_path / "parts" / part_name
        part_dir.mkdir(parents=True)
        with (part_dir / "state").open("w") as state_file:
            print(step.name, file=state_file, end="")

        handler = unit.load_part(part_name)

        assert handler.latest_step() == step



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
from textwrap import dedent

from testtools.matchers import Contains, FileExists, Not

from tests import integration


class StagePackageGrammarTestCase(integration.TestCase):
    def test_simple(self):
        """Test that 'simple' fetches stage package."""

        self.run_snapcraft(["prime", "simple"], "stage-package-grammar")

        self.assertThat(os.path.join("prime", "usr", "bin", "hello"), FileExists())

    def test_try(self):
        """Test that 'try' fetches stage package."""

        self.run_snapcraft(["prime", "try"], "stage-package-grammar")

        self.assertThat(os.path.join("prime", "usr", "bin", "hello"), FileExists())

    def test_try_skipped(self):
        """Test that 'try-skipped' fetches nothing."""

        self.run_snapcraft(["prime", "try-skipped"], "stage-package-grammar")

        self.assertThat(os.path.join("prime", "usr", "bin", "hello"), Not(FileExists()))

    def test_try_else(self):
        """Test that 'try-else' fetches stage package."""

        self.run_snapcraft(["prime", "try-else"], "stage-package-grammar")

        self.assertThat(os.path.join("prime", "usr", "bin", "hello"), FileExists())

    def test_on_other_arch(self):
        """Test that 'on-other-arch' fetches nothing."""

        self.run_snapcraft(["prime", "on-other-arch"], "stage-package-grammar")

        self.assertThat(os.path.join("prime", "usr", "bin", "hello"), Not(FileExists()))

    def test_on_other_arch_else(self):
        """Test that 'on-other-arch-else' fetches stage package."""

        self.run_snapcraft(["prime", "on-other-arch-else"], "stage-package-grammar")

        self.assertThat(os.path.join("prime", "usr", "bin", "hello"), FileExists())

    def test_on_other_arch_else_fail(self):
        """Test that 'on-other-arch-else-fail' fails with an error."""

        exception = self.assertRaises(
            subprocess.CalledProcessError,
            self.run_snapcraft,
            ["prime", "on-other-arch-else-fail"],
            "stage-package-grammar",
        )

        self.assertThat(
            exception.output,
            Contains("Unable to satisfy 'on other-arch', failure forced"),
        )

    def test_to_other_arch(self):
        """Test that 'to' for the other arch fetches nothing."""

        self.construct_yaml(
            parts=dedent(
                """\
            simple:
              plugin: nil
              stage-packages:
              - to other-arch:
                - hello
            """
            )
        )
        self.run_snapcraft(["prime", "simple"])
        self.assertThat(os.path.join("prime", "usr", "bin", "hello"), Not(FileExists()))

    def test_to_other_arch_else(self):
        """Test that 'else' for the other arch fetches hello."""

        self.construct_yaml(
            parts=dedent(
                """\
            simple:
              plugin: nil
              stage-packages:
              - to other-arch:
                - foo
              - else:
                - hello
            """
            )
        )
        self.run_snapcraft(["prime", "simple"])
        self.assertThat(os.path.join("prime", "usr", "bin", "hello"), FileExists())

    def test_to_other_arch_else_fail(self):
        """Test that 'else' for the other arch fails."""

        self.construct_yaml(
            parts=dedent(
                """\
            simple:
              plugin: nil
              stage-packages:
              - to other-arch:
                - foo
              - else fail
            """
            )
        )
        exception = self.assertRaises(
            subprocess.CalledProcessError, self.run_snapcraft, ["prime", "simple"]
        )

        self.assertThat(
            exception.output,
            Contains("Unable to satisfy 'to other-arch', failure forced"),
        )

    def test_on_to_other_arch(self):
        """Test that 'on to' for the other arch fetches nothing."""

        self.construct_yaml(
            parts=dedent(
                """\
            simple:
              plugin: nil
              stage-packages:
              - on i386 to other-arch:
                - hello
            """
            )
        )
        self.run_snapcraft(["prime", "simple"])
        self.assertThat(os.path.join("prime", "usr", "bin", "hello"), Not(FileExists()))

    def test_on_to_other_arch_else(self):
        """Test that 'else' for the other arch fetches hello."""

        self.construct_yaml(
            parts=dedent(
                """\
            simple:
              plugin: nil
              stage-packages:
              - on i386 to other-arch:
                - foo
              - else:
                - hello
            """
            )
        )
        self.run_snapcraft(["prime", "simple"])
        self.assertThat(os.path.join("prime", "usr", "bin", "hello"), FileExists())

    def test_on_to_other_arch_else_fail(self):
        """Test that 'else' for the other arch fails."""

        self.construct_yaml(
            parts=dedent(
                """\
            simple:
              plugin: nil
              stage-packages:
              - on i386 to other-arch:
                - foo
              - else fail
            """
            )
        )
        exception = self.assertRaises(
            subprocess.CalledProcessError, self.run_snapcraft, ["prime", "simple"]
        )

        self.assertThat(
            exception.output,
            Contains("Unable to satisfy 'on i386 to other-arch', failure forced"),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import sys
import tempfile
from textwrap import dedent
from unittest import mock

import fixtures
from testtools.matchers import Contains, Equals, GreaterThan, Not

import snapcraft
from snapcraft.internal import common
from tests.fixture_setup.os_release import FakeOsRelease

from . import ProjectLoaderBaseTest


class EnvironmentTest(ProjectLoaderBaseTest):
    def setUp(self):
        super().setUp()

        self.snapcraft_yaml = dedent(
            """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: test
            confinement: strict
            grade: stable

            parts:
              part1:
                plugin: nil
        """
        )

    def test_config_snap_environment(self):
        project_config = self.make_snapcraft_project(self.snapcraft_yaml)

        lib_paths = [
            os.path.join(self.prime_dir, "lib"),
            os.path.join(self.prime_dir, "usr", "lib"),
        ]
        for lib_path in lib_paths:
            os.makedirs(lib_path)

        environment = project_config.snap_env()
        self.assertThat(
            environment,
            Contains(
                'PATH="{0}/usr/sbin:{0}/usr/bin:{0}/sbin:{0}/bin${{PATH:+:$PATH}}"'.format(
                    self.prime_dir
                )
            ),
        )
        self.assertThat(
            environment,
            Contains(
                'LD_LIBRARY_PATH="${{LD_LIBRARY_PATH:+$LD_LIBRARY_PATH:}}'
                '{0}/lib:{0}/usr/lib"'.format(self.prime_dir)
            ),
        )

    def test_config_snap_environment_with_no_library_paths(self):
        project_config = self.make_snapcraft_project(self.snapcraft_yaml)

        environment = project_config.snap_env()
        self.assertTrue(
            'PATH="{0}/usr/sbin:{0}/usr/bin:{0}/sbin:{0}/bin${{PATH:+:$PATH}}"'.format(
                self.prime_dir
            )
            in environment,
            "Current PATH is {!r}".format(environment),
        )
        for e in environment:
            self.assertFalse(
                "LD_LIBRARY_PATH" in e, "Current environment is {!r}".format(e)
            )

    @mock.patch.object(
        snapcraft.internal.pluginhandler.PluginHandler, "get_primed_dependency_paths"
    )
    def test_config_snap_environment_with_dependencies(self, mock_get_dependencies):
        library_paths = {
            os.path.join(self.prime_dir, "lib1"),
            os.path.join(self.prime_dir, "lib2"),
        }
        mock_get_dependencies.return_value = library_paths
        project_config = self.make_snapcraft_project(self.snapcraft_yaml)

        for lib_path in library_paths:
            os.makedirs(lib_path)

        # Ensure that LD_LIBRARY_PATH is present and it contains the
        # extra dependency paths.
        self.assertThat(
            project_config.snap_env(),
            Contains(
                'LD_LIBRARY_PATH="{0}/lib1:{0}/lib2${{LD_LIBRARY_PATH:+:$LD_LIBRARY_PATH}}"'.format(
                    self.prime_dir
                )
            ),
        )

    @mock.patch.object(
        snapcraft.internal.pluginhandler.PluginHandler, "get_primed_dependency_paths"
    )
    def test_config_snap_environment_with_dependencies_but_no_paths(
        self, mock_get_dependencies
    ):
        library_paths = {
            os.path.join(self.prime_dir, "lib1"),
            os.path.join(self.prime_dir, "lib2"),
        }
        mock_get_dependencies.return_value = library_paths
        project_config = self.make_snapcraft_project(self.snapcraft_yaml)

        # Ensure that LD_LIBRARY_PATH is present, but is completey empty since
        # no library paths actually exist.
        for variable in project_config.snap_env():
            self.assertFalse(
                "LD_LIBRARY_PATH" in variable,
                "Expected no LD_LIBRARY_PATH (got {!r})".format(variable),
            )

    def test_config_runtime_environment_ld(self):
        # Place a few ld.so.conf files in supported locations. We expect the
        # contents of these to make it into the LD_LIBRARY_PATH.
        mesa_dir = os.path.join(self.prime_dir, "usr", "lib", "my_arch", "mesa")
        os.makedirs(mesa_dir)
        with open(os.path.join(mesa_dir, "ld.so.conf"), "w") as f:
            f.write("/mesa")

        mesa_egl_dir = os.path.join(self.prime_dir, "usr", "lib", "my_arch", "mesa-egl")
        os.makedirs(mesa_egl_dir)
        with open(os.path.join(mesa_egl_dir, "ld.so.conf"), "w") as f:
            f.write("# Standalone comment\n")
            f.write("/mesa-egl")

        project_config = self.make_snapcraft_project(self.snapcraft_yaml)
        environment = project_config.snap_env()

        # Ensure that the LD_LIBRARY_PATH includes all the above paths
        paths = []
        for variable in environment:
            if "LD_LIBRARY_PATH" in variable:
                these_paths = variable.split("=")[1].strip()
                paths.extend(these_paths.replace('"', "").split(":"))

        self.assertTrue(len(paths) > 0, "Expected LD_LIBRARY_PATH to be in environment")

        expected = (os.path.join(self.prime_dir, i) for i in ["mesa", "mesa-egl"])
        for item in expected:
            self.assertTrue(
                item in paths, 'Expected LD_LIBRARY_PATH to include "{}"'.format(item)
            )

    def test_config_env_dedup(self):
        """Regression test for LP: #1767625.
        Verify that the use of after with multiple parts does not produce
        duplicate exports.
        """
        snapcraft_yaml = dedent(
            """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: test
            confinement: strict
            grade: stable

            parts:
              main:
                plugin: nil
                after: [part1, part2, part3]
              part1:
                plugin: nil
              part2:
                plugin: nil
              part3:
                plugin: nil
        """
        )
        project_config = self.make_snapcraft_project(snapcraft_yaml)
        part = project_config.parts.get_part("main")
        environment = project_config.parts.build_env_for_part(part, root_part=True)
        # We sort here for equality checking but they should not be sorted
        # for a real case scenario.
        environment.sort()
        self.assertThat(
            environment,
            Equals(
                [
                    (
                        'PATH="{0}/parts/main/install/usr/sbin:'
                        "{0}/parts/main/install/usr/bin:"
                        "{0}/parts/main/install/sbin:"
                        '{0}/parts/main/install/bin${{PATH:+:$PATH}}"'
                    ).format(self.path),
                    (
                        'PATH="{0}/stage/usr/sbin:'
                        "{0}/stage/usr/bin:"
                        "{0}/stage/sbin:"
                        '{0}/stage/bin${{PATH:+:$PATH}}"'
                    ).format(self.path),
                    'PERL5LIB="{0}/stage/usr/share/perl5/"'.format(self.path),
                    'SNAPCRAFT_ARCH_TRIPLET="{}"'.format(
                        project_config.project.arch_triplet
                    ),
                    'SNAPCRAFT_EXTENSIONS_DIR="{}"'.format(common.get_extensionsdir()),
                    'SNAPCRAFT_PARALLEL_BUILD_COUNT="2"',
                    'SNAPCRAFT_PART_BUILD="{}/parts/main/build"'.format(self.path),
                    'SNAPCRAFT_PART_BUILD_WORK="{}/parts/main/build/"'.format(
                        self.path
                    ),
                    'SNAPCRAFT_PART_INSTALL="{}/parts/main/install"'.format(self.path),
                    'SNAPCRAFT_PART_SRC="{}/parts/main/src"'.format(self.path),
                    'SNAPCRAFT_PART_SRC_WORK="{}/parts/main/src/"'.format(self.path),
                    'SNAPCRAFT_PRIME="{}/prime"'.format(self.path),
                    'SNAPCRAFT_PROJECT_DIR="{}"'.format(self.path),
                    'SNAPCRAFT_PROJECT_GRADE="stable"',
                    'SNAPCRAFT_PROJECT_NAME="test"',
                    'SNAPCRAFT_PROJECT_VERSION="1"',
                    'SNAPCRAFT_STAGE="{}/stage"'.format(self.path),
                    'SNAPCRAFT_TARGET_ARCH="{}"'.format(
                        project_config.project.target_arch
                    ),
                ]
            ),
        )

    def test_config_stage_environment_confinement_classic(self):
        self.useFixture(FakeOsRelease())

        snapcraft_yaml = dedent(
            """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: test
            confinement: classic
            grade: stable
            base: core

            parts:
              part1:
                plugin: nil
        """
        )
        project_config = self.make_snapcraft_project(snapcraft_yaml)
        part = project_config.parts.get_part("part1")
        environment = project_config.parts.build_env_for_part(part, root_part=True)
        self.assertThat(
            environment,
            Not(
                Contains(
                    'LD_LIBRARY_PATH="$LD_LIBRARY_PATH:{base_core_path}/lib:'
                    "{base_core_path}/usr/lib:{base_core_path}/lib/{arch_triplet}:"
                    '{base_core_path}/usr/lib/{arch_triplet}"'.format(
                        base_core_path=self.base_environment.core_path,
                        arch_triplet=project_config.project.arch_triplet,
                    )
                )
            ),
        )

    def test_parts_build_env_ordering_with_deps(self):
        snapcraft_yaml = dedent(
            """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: test
            confinement: strict
            grade: stable

            parts:
              part1:
                plugin: nil
              part2:
                plugin: nil
                after: [part1]
        """
        )

        self.useFixture(fixtures.EnvironmentVariable("PATH", "/bin"))

        arch_triplet = snapcraft.ProjectOptions().arch_triplet
        self.maxDiff = None
        paths = [
            os.path.join(self.stage_dir, "lib"),
            os.path.join(self.stage_dir, "lib", arch_triplet),
            os.path.join(self.stage_dir, "usr", "lib"),
            os.path.join(self.stage_dir, "usr", "lib", arch_triplet),
            os.path.join(self.stage_dir, "include"),
            os.path.join(self.stage_dir, "usr", "include"),
            os.path.join(self.stage_dir, "include", arch_triplet),
            os.path.join(self.stage_dir, "usr", "include", arch_triplet),
            os.path.join(self.parts_dir, "part1", "install", "include"),
            os.path.join(self.parts_dir, "part1", "install", "lib"),
            os.path.join(self.parts_dir, "part2", "install", "include"),
            os.path.join(self.parts_dir, "part2", "install", "lib"),
        ]
        for path in paths:
            os.makedirs(path)

        project_config = self.make_snapcraft_project(snapcraft_yaml)
        part2 = [
            part for part in project_config.parts.all_parts if part.name == "part2"
        ][0]
        env = project_config.parts.build_env_for_part(part2)
        env_lines = "\n".join(["export {}\n".format(e) for e in env])

        shell_env = {
            "CFLAGS": "-I/user-provided",
            "CXXFLAGS": "-I/user-provided",
            "CPPFLAGS": "-I/user-provided",
            "LDFLAGS": "-L/user-provided",
            "LD_LIBRARY_PATH": "/user-provided",
        }

        def get_envvar(envvar):
            with tempfile.NamedTemporaryFile(mode="w+") as f:
                f.write(env_lines)
                f.write("echo ${}".format(envvar))
                f.flush()
                output = subprocess.check_output(["/bin/sh", f.name], env=shell_env)
            return output.decode(sys.getfilesystemencoding()).strip()

        expected_cflags = (
            "-I/user-provided "
            "-isystem{parts_dir}/part2/install/include -isystem{stage_dir}/include "
            "-isystem{stage_dir}/usr/include "
            "-isystem{stage_dir}/include/{arch_triplet} "
            "-isystem{stage_dir}/usr/include/{arch_triplet}".format(
                parts_dir=self.parts_dir,
                stage_dir=self.stage_dir,
                arch_triplet=project_config.project.arch_triplet,
            )
        )
        self.assertThat(get_envvar("CFLAGS"), Equals(expected_cflags))
        self.assertThat(get_envvar("CXXFLAGS"), Equals(expected_cflags))
        self.assertThat(get_envvar("CPPFLAGS"), Equals(expected_cflags))

        self.assertThat(
            get_envvar("LDFLAGS"),
            Equals(
                "-L/user-provided "
                "-L{parts_dir}/part2/install/lib -L{stage_dir}/lib "
                "-L{stage_dir}/usr/lib -L{stage_dir}/lib/{arch_triplet} "
                "-L{stage_dir}/usr/lib/{arch_triplet}".format(
                    parts_dir=self.parts_dir,
                    stage_dir=self.stage_dir,
                    arch_triplet=project_config.project.arch_triplet,
                )
            ),
        )

        self.assertThat(
            get_envvar("LD_LIBRARY_PATH"),
            Equals(
                "/user-provided:"
                "{parts_dir}/part2/install/lib:"
                "{stage_dir}/lib:"
                "{stage_dir}/usr/lib:"
                "{stage_dir}/lib/{arch_triplet}:"
                "{stage_dir}/usr/lib/{arch_triplet}".format(
                    parts_dir=self.parts_dir,
                    stage_dir=self.stage_dir,
                    arch_triplet=project_config.project.arch_triplet,
                )
            ),
        )

    @mock.patch("os.sched_getaffinity", return_value=set(range(0, 42)))
    def test_parts_build_env_contains_parallel_build_count(self, cpu_mock):
        project_config = self.make_snapcraft_project(self.snapcraft_yaml)
        part1 = [
            part for part in project_config.parts.all_parts if part.name == "part1"
        ][0]
        env = project_config.parts.build_env_for_part(part1)
        self.assertThat(env, Contains('SNAPCRAFT_PARALLEL_BUILD_COUNT="42"'))

    @mock.patch("os.sched_getaffinity", side_effect=AttributeError)
    @mock.patch("multiprocessing.cpu_count", return_value=42)
    def test_parts_build_env_contains_parallel_build_count_no_getaffinity(
        self, affinity_mock, cpu_mock
    ):
        project_config = self.make_snapcraft_project(self.snapcraft_yaml)
        part1 = [
            part for part in project_config.parts.all_parts if part.name == "part1"
        ][0]
        env = project_config.parts.build_env_for_part(part1)
        self.assertThat(env, Contains('SNAPCRAFT_PARALLEL_BUILD_COUNT="42"'))

    @mock.patch("os.sched_getaffinity", side_effect=AttributeError)
    @mock.patch("multiprocessing.cpu_count", side_effect=NotImplementedError)
    def test_parts_build_env_contains_parallel_build_count_no_cpucount(
        self, affinity_mock, cpu_mock
    ):
        project_config = self.make_snapcraft_project(self.snapcraft_yaml)
        part1 = [
            part for part in project_config.parts.all_parts if part.name == "part1"
        ][0]
        env = project_config.parts.build_env_for_part(part1)
        self.assertThat(env, Contains('SNAPCRAFT_PARALLEL_BUILD_COUNT="1"'))

    def test_extension_dir(self):
        common.set_extensionsdir("/foo")
        project_config = self.make_snapcraft_project(self.snapcraft_yaml)
        part1 = [
            part for part in project_config.parts.all_parts if part.name == "part1"
        ][0]
        env = project_config.parts.build_env_for_part(part1)
        self.assertThat(env, Contains('SNAPCRAFT_EXTENSIONS_DIR="/foo"'))

    def test_project_dir(self):
        project_config = self.make_snapcraft_project(self.snapcraft_yaml)
        env = project_config.parts.build_env_for_part(project_config.parts.all_parts[0])
        self.assertThat(env, Contains('SNAPCRAFT_PROJECT_DIR="{}"'.format(self.path)))

    def test_build_environment(self):
        self.useFixture(FakeOsRelease())

        snapcraft_yaml = dedent(
            """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: test
            confinement: strict
            grade: stable
            base: core

            parts:
              part1:
                plugin: nil
                build-environment:
                  - FOO: BAR
        """
        )
        project_config = self.make_snapcraft_project(snapcraft_yaml)
        part = project_config.parts.get_part("part1")
        environment = project_config.parts.build_env_for_part(part)
        self.assertThat(environment, Contains('FOO="BAR"'))

    def test_build_environment_can_depend_on_global_env(self):
        self.useFixture(FakeOsRelease())

        snapcraft_yaml = dedent(
            """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: test
            confinement: strict
            grade: stable
            base: core

            parts:
              part1:
                plugin: nil
                build-environment:
                  - PROJECT_NAME: $SNAPCRAFT_PROJECT_NAME
        """
        )
        project_config = self.make_snapcraft_project(snapcraft_yaml)
        part = project_config.parts.get_part("part1")
        environment = project_config.parts.build_env_for_part(part)
        snapcraft_definition_index = -1
        build_environment_definition_index = -1
        for index, variable in enumerate(environment):
            if variable.startswith("SNAPCRAFT_PROJECT_NAME="):
                snapcraft_definition_index = index
            if variable.startswith("PROJECT_NAME="):
                build_environment_definition_index = index

        # Assert that each definition was found, and the global env came before the
        # build environment
        self.assertThat(snapcraft_definition_index, GreaterThan(-1))
        self.assertThat(build_environment_definition_index, GreaterThan(-1))
        self.assertThat(
            build_environment_definition_index, GreaterThan(snapcraft_definition_index)
        )

    def test_build_environment_can_depend_on_part_env(self):
        self.useFixture(FakeOsRelease())

        snapcraft_yaml = dedent(
            """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: test
            confinement: strict
            grade: stable
            base: core

            parts:
              part1:
                plugin: nil
                build-environment:
                  - PART_INSTALL: $SNAPCRAFT_PART_INSTALL
        """
        )
        project_config = self.make_snapcraft_project(snapcraft_yaml)
        part = project_config.parts.get_part("part1")
        environment = project_config.parts.build_env_for_part(part)
        snapcraft_definition_index = -1
        build_environment_definition_index = -1
        for index, variable in enumerate(environment):
            if variable.startswith("SNAPCRAFT_PART_INSTALL="):
                snapcraft_definition_index = index
            if variable.startswith("PART_INSTALL="):
                build_environment_definition_index = index

        # Assert that each definition was found, and the part env came before the
        # build environment
        self.assertThat(snapcraft_definition_index, GreaterThan(-1))
        self.assertThat(build_environment_definition_index, GreaterThan(-1))
        self.assertThat(
            build_environment_definition_index, GreaterThan(snapcraft_definition_index)
        )

    def test_build_environment_with_dependencies_does_not_leak(self):
        self.useFixture(FakeOsRelease())

        snapcraft_yaml = dedent(
            """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: test
            confinement: strict
            grade: stable
            base: core

            parts:
              part1:
                plugin: nil
                build-environment:
                  - FOO: BAR

              part2:
                plugin: nil
                after: [part1]
                build-environment:
                  - BAZ: QUX
        """
        )
        project_config = self.make_snapcraft_project(snapcraft_yaml)
        part1 = project_config.parts.get_part("part1")
        part2 = project_config.parts.get_part("part2")
        self.assertThat(
            project_config.parts.build_env_for_part(part1), Contains('FOO="BAR"')
        )
        self.assertThat(
            project_config.parts.build_env_for_part(part2), Not(Contains('FOO="BAR"'))
        )
        self.assertThat(
            project_config.parts.build_env_for_part(part2), Contains('BAZ="QUX"')
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Manage the host's apt source repository configuration."""

import io
import logging
import os
import pathlib
import re
import subprocess
import tempfile
from typing import List, Optional

from snapcraft.internal import os_release
from snapcraft.internal.meta import package_repository
from snapcraft.project._project_options import ProjectOptions

from . import apt_ppa

logger = logging.getLogger(__name__)


def _construct_deb822_source(
    *,
    architectures: Optional[List[str]] = None,
    components: Optional[List[str]] = None,
    formats: Optional[List[str]] = None,
    suites: List[str],
    url: str,
) -> str:
    """Construct deb-822 formatted sources.list config string."""
    with io.StringIO() as deb822:
        if formats:
            type_text = " ".join(formats)
        else:
            type_text = "deb"

        print(f"Types: {type_text}", file=deb822)

        print(f"URIs: {url}", file=deb822)

        suites_text = " ".join(suites)
        print(f"Suites: {suites_text}", file=deb822)

        if components:
            components_text = " ".join(components)
            print(f"Components: {components_text}", file=deb822)

        if architectures:
            arch_text = " ".join(architectures)
        else:
            arch_text = _get_host_arch()

        print(f"Architectures: {arch_text}", file=deb822)

        return deb822.getvalue()


def _get_host_arch() -> str:
    return ProjectOptions().deb_arch


def _sudo_write_file(*, dst_path: pathlib.Path, content: bytes) -> None:
    """Workaround for writing privileged files in destructive mode."""
    try:
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            temp_file.write(content)
            temp_file.flush()
            f_name = temp_file.name

        try:
            command = [
                "sudo",
                "install",
                "--owner=root",
                "--group=root",
                "--mode=0644",
                f_name,
                str(dst_path),
            ]
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError as error:
            raise RuntimeError(
                f"Failed to install repository config with: {command!r}"
            ) from error
    finally:
        os.unlink(f_name)


class AptSourcesManager:
    """Manage apt source configuration in /etc/apt/sources.list.d.

    :param sources_list_d: Path to sources.list.d directory.
    """

    # pylint: disable=too-few-public-methods
    def __init__(
        self, *, sources_list_d: pathlib.Path = pathlib.Path("/etc/apt/sources.list.d"),
    ) -> None:
        self._sources_list_d = sources_list_d

    def _install_sources(
        self,
        *,
        architectures: Optional[List[str]] = None,
        components: Optional[List[str]] = None,
        formats: Optional[List[str]] = None,
        name: str,
        suites: List[str],
        url: str,
    ) -> bool:
        """Install sources list configuration.

        Write config to:
        /etc/apt/sources.list.d/snapcraft-<name>.sources

        :returns: True if configuration was changed.
        """
        config = _construct_deb822_source(
            architectures=architectures,
            components=components,
            formats=formats,
            suites=suites,
            url=url,
        )

        if name not in ["default", "default-security"]:
            name = "snapcraft-" + name

        config_path = self._sources_list_d / f"{name}.sources"
        if config_path.exists() and config_path.read_text() == config:
            # Already installed and matches, nothing to do.
            logger.debug("Ignoring unchanged sources: %s", str(config_path))
            return False

        _sudo_write_file(dst_path=config_path, content=config.encode())
        logger.debug("Installed sources: %s", str(config_path))
        return True

    def _install_sources_apt(
        self, *, package_repo: package_repository.PackageRepositoryApt
    ) -> bool:
        """Install repository configuration.

        1) First check to see if package repo is implied path,
           or "bare repository" config.  This is indicated when no
           path, components, or suites are indicated.
        2) If path is specified, convert path to a suite entry,
           ending with "/".

        Relatedly, this assumes all of the error-checking has been
        done already on the package_repository object in a proper
        fashion, but do some sanity checks here anyways.

        :returns: True if source configuration was changed.
        """
        if (
            not package_repo.path
            and not package_repo.components
            and not package_repo.suites
        ):
            suites = ["/"]
        elif package_repo.path:
            # Suites denoting exact path must end with '/'.
            path = package_repo.path
            if not path.endswith("/"):
                path += "/"
            suites = [path]
        elif package_repo.suites:
            suites = package_repo.suites
            if not package_repo.components:
                raise RuntimeError("no components with suite")
        else:
            raise RuntimeError("no suites or path")

        if package_repo.name:
            name = package_repo.name
        else:
            name = re.sub(r"\W+", "_", package_repo.url)

        return self._install_sources(
            architectures=package_repo.architectures,
            components=package_repo.components,
            formats=package_repo.formats,
            name=name,
            suites=suites,
            url=package_repo.url,
        )

    def _install_sources_ppa(
        self, *, package_repo: package_repository.PackageRepositoryAptPpa
    ) -> bool:
        """Install PPA formatted repository.

        Create a sources list config by:
        - Looking up the codename of the host OS and using it as the "suites"
          entry.
        - Formulate deb URL to point to PPA.
        - Enable only "deb" formats.

        :returns: True if source configuration was changed.
        """
        owner, name = apt_ppa.split_ppa_parts(ppa=package_repo.ppa)
        codename = os_release.OsRelease().version_codename()

        return self._install_sources(
            components=["main"],
            formats=["deb"],
            name=f"ppa-{owner}_{name}",
            suites=[codename],
            url=f"http://ppa.launchpad.net/{owner}/{name}/ubuntu",
        )

    def install_package_repository_sources(
        self, *, package_repo: package_repository.PackageRepository,
    ) -> bool:
        """Install configured package repositories.

        :param package_repo: Repository to install the source configuration for.

        :returns: True if source configuration was changed.
        """
        logger.debug("Processing repo: %r", package_repo)
        if isinstance(package_repo, package_repository.PackageRepositoryAptPpa):
            return self._install_sources_ppa(package_repo=package_repo)

        if isinstance(package_repo, package_repository.PackageRepositoryApt):
            return self._install_sources_apt(package_repo=package_repo)

        raise RuntimeError(f"unhandled package repository: {package_repository!r}")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

from tests import unit


class SourceTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        patcher = mock.patch("subprocess.check_call")
        self.mock_run = patcher.start()
        self.mock_run.return_value = True
        self.addCleanup(patcher.stop)

        patcher = mock.patch("os.rmdir")
        self.mock_rmdir = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("os.path.exists")
        self.mock_path_exists = patcher.start()
        self.mock_path_exists.return_value = False
        self.addCleanup(patcher.stop)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from testtools.matchers import Equals

from snapcraft.internal import project_loader
from tests import unit


class VariableReplacementsTest(unit.TestCase):
    def test_string_replacements(self):
        replacements = (
            ("no replacement", "snapcraft_stage/usr/bin", "snapcraft_stage/usr/bin"),
            (
                "replaced start",
                "$SNAPCRAFT_STAGE/usr/bin",
                "{}/usr/bin".format(self.stage_dir),
            ),
            (
                "replaced between",
                "--with-swig $SNAPCRAFT_STAGE/usr/swig",
                "--with-swig {}/usr/swig".format(self.stage_dir),
            ),
            (
                "project replacement",
                "$SNAPCRAFT_PROJECT_NAME-$SNAPCRAFT_PROJECT_VERSION",
                "project_name-version",
            ),
            (
                "multiple replacement",
                "$SNAPCRAFT_PROJECT_NAME-$SNAPCRAFT_PROJECT_NAME",
                "project_name-project_name",
            ),
        )

        for test_name, subject, expected in replacements:
            self.assertThat(
                project_loader.replace_attr(
                    subject,
                    {
                        "$SNAPCRAFT_PROJECT_NAME": "project_name",
                        "$SNAPCRAFT_PROJECT_VERSION": "version",
                        "$SNAPCRAFT_STAGE": self.stage_dir,
                    },
                ),
                Equals(expected),
            )

    def test_lists_with_string_replacements(self):
        replacements = (
            (
                "no replacement",
                ["snapcraft_stage/usr/bin", "/usr/bin"],
                ["snapcraft_stage/usr/bin", "/usr/bin"],
            ),
            (
                "replaced start",
                ["$SNAPCRAFT_STAGE/usr/bin", "/usr/bin"],
                ["{}/usr/bin".format(self.stage_dir), "/usr/bin"],
            ),
            (
                "replaced between",
                ["--without-python", "--with-swig $SNAPCRAFT_STAGE/usr/swig"],
                ["--without-python", "--with-swig {}/usr/swig".format(self.stage_dir)],
            ),
        )

        for test_name, subject, expected in replacements:
            self.assertThat(
                project_loader.replace_attr(
                    subject,
                    {
                        "$SNAPCRAFT_PROJECT_NAME": "project_name",
                        "$SNAPCRAFT_PROJECT_VERSION": "version",
                        "$SNAPCRAFT_STAGE": self.stage_dir,
                    },
                ),
                Equals(expected),
            )

    def test_tuples_with_string_replacements(self):
        replacements = (
            (
                "no replacement",
                ("snapcraft_stage/usr/bin", "/usr/bin"),
                ["snapcraft_stage/usr/bin", "/usr/bin"],
            ),
            (
                "replaced start",
                ("$SNAPCRAFT_STAGE/usr/bin", "/usr/bin"),
                ["{}/usr/bin".format(self.stage_dir), "/usr/bin"],
            ),
            (
                "replaced between",
                ("--without-python", "--with-swig $SNAPCRAFT_STAGE/usr/swig"),
                ["--without-python", "--with-swig {}/usr/swig".format(self.stage_dir)],
            ),
        )

        for test_name, subject, expected in replacements:
            self.assertThat(
                project_loader.replace_attr(
                    subject,
                    {
                        "$SNAPCRAFT_PROJECT_NAME": "project_name",
                        "$SNAPCRAFT_PROJECT_VERSION": "version",
                        "$SNAPCRAFT_STAGE": self.stage_dir,
                    },
                ),
                Equals(expected),
            )

    def test_dict_with_string_replacements(self):
        replacements = (
            (
                "no replacement",
                {"1": "snapcraft_stage/usr/bin", "2": "/usr/bin"},
                {"1": "snapcraft_stage/usr/bin", "2": "/usr/bin"},
            ),
            (
                "replaced start",
                {"1": "$SNAPCRAFT_STAGE/usr/bin", "2": "/usr/bin"},
                {"1": "{}/usr/bin".format(self.stage_dir), "2": "/usr/bin"},
            ),
            (
                "replaced between",
                {"1": "--without-python", "2": "--with-swig $SNAPCRAFT_STAGE/usr/swig"},
                {
                    "1": "--without-python",
                    "2": "--with-swig {}/usr/swig".format(self.stage_dir),
                },
            ),
            (
                "replace keys as well",
                {"$SNAPCRAFT_STAGE": "--with-swig $SNAPCRAFT_STAGE"},
                {self.stage_dir: "--with-swig {}".format(self.stage_dir)},
            ),
        )

        for test_name, subject, expected in replacements:
            self.assertThat(
                project_loader.replace_attr(
                    subject,
                    {
                        "$SNAPCRAFT_PROJECT_NAME": "project_name",
                        "$SNAPCRAFT_PROJECT_VERSION": "version",
                        "$SNAPCRAFT_STAGE": self.stage_dir,
                    },
                ),
                Equals(expected),
            )

    def test_string_replacement_with_complex_data(self):
        subject = {
            "filesets": {
                "files": [
                    "somefile",
                    "$SNAPCRAFT_STAGE/file1",
                    "SNAPCRAFT_STAGE/really",
                ]
            },
            "configflags": ["--with-python", "--with-swig $SNAPCRAFT_STAGE/swig"],
        }

        expected = {
            "filesets": {
                "files": [
                    "somefile",
                    "{}/file1".format(self.stage_dir),
                    "SNAPCRAFT_STAGE/really",
                ]
            },
            "configflags": [
                "--with-python",
                "--with-swig {}/swig".format(self.stage_dir),
            ],
        }

        self.assertThat(
            project_loader.replace_attr(
                subject,
                {
                    "$SNAPCRAFT_PROJECT_NAME": "project_name",
                    "$SNAPCRAFT_PROJECT_VERSION": "version",
                    "$SNAPCRAFT_STAGE": self.stage_dir,
                },
            ),
            Equals(expected),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2017, 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import shlex
from typing import List

from snapcraft import formatting_utils
from snapcraft.internal import errors


class SnapcraftSourceError(errors.SnapcraftError):
    pass


class VCSError(SnapcraftSourceError):
    fmt = "{message}"


class SnapcraftSourceNotFoundError(SnapcraftSourceError):

    fmt = (
        "Failed to pull source: {source!r}.\n"
        "Please ensure the source path is correct and that it is accessible.\n"
        "See `snapcraft help sources` for more information."
    )

    def __init__(self, source):
        super().__init__(source=source)


class SnapcraftSourceUnhandledError(SnapcraftSourceError):

    fmt = (
        "Failed to pull source: "
        "unable to determine source type of {source!r}.\n"
        "Check that the URL is correct or "
        "consider specifying `source-type` for this part. "
        "See `snapcraft help sources` for more information."
    )

    def __init__(self, source):
        super().__init__(source=source)


class SnapcraftSourceInvalidOptionError(SnapcraftSourceError):

    fmt = (
        "Failed to pull source: "
        "{option!r} cannot be used with a {source_type} source.\n"
        "See `snapcraft help sources` for more information."
    )

    def __init__(self, source_type: str, option: str) -> None:
        super().__init__(source_type=source_type, option=option)


class SnapcraftSourceIncompatibleOptionsError(SnapcraftSourceError):

    fmt = (
        "Failed to pull source: "
        "cannot specify both {humanized_options} for a {source_type} source.\n"
        "See `snapcraft help sources` for more information."
    )

    def __init__(self, source_type: str, options: List[str]) -> None:
        self.options = options
        super().__init__(
            source_type=source_type,
            humanized_options=formatting_utils.humanize_list(options, "and"),
        )


class DigestDoesNotMatchError(SnapcraftSourceError):

    fmt = "Expected the digest for source to be {expected}, but it was {calculated}"

    def __init__(self, expected, calculated):
        super().__init__(expected=expected, calculated=calculated)


class InvalidDebError(SnapcraftSourceError):

    fmt = (
        "The {deb_file} used does not contain valid data. "
        "Ensure a proper deb file is passed for .deb files "
        "as sources."
    )


class InvalidSnapError(SnapcraftSourceError):

    fmt = (
        "The snap file does not contain valid data. "
        "Ensure the source lists a proper snap file"
    )


class SourceUpdateUnsupportedError(SnapcraftSourceError):

    fmt = "Failed to update source: {source!s} sources don't support updating."

    def __init__(self, source):
        super().__init__(source=source)


class SnapcraftPullError(SnapcraftSourceError):

    fmt = "Failed to pull source, command {command!r} exited with code {exit_code}."

    def __init__(self, command, exit_code):
        if isinstance(command, list):
            string_command = " ".join(shlex.quote(i) for i in command)
        else:
            string_command = command
        super().__init__(command=string_command, exit_code=exit_code)


class SnapcraftRequestError(SnapcraftSourceError):

    fmt = "Network request error: {message}"


class GitCommandError(errors.SnapcraftException):
    def __init__(self, *, command: List[str], exit_code: int, output: str) -> None:
        self._command = " ".join(shlex.quote(i) for i in command)
        self._exit_code = exit_code
        self._output = output

    def get_brief(self) -> str:
        return f"Failed to execute git command: {self._command}"

    def get_details(self) -> str:
        return f"Command failed with exit code {self._exit_code!r} and output:\n{self._output}"

    def get_resolution(self) -> str:
        return "Consider checking your git configuration for settings which may cause issues."



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

import yaml as yaml_utils
from testtools.matchers import Equals

from tests import integration


class ArchitecturesTestCase(integration.TestCase):
    def test_simple_architecture(self):
        self.construct_yaml(architectures=[self.deb_arch])

        self.run_snapcraft("prime")

        with open(os.path.join("prime", "meta", "snap.yaml")) as f:
            y = yaml_utils.load(f)

        self.assertThat(y["architectures"], Equals([self.deb_arch]))

    def test_build_and_run_on_architecture(self):
        self.construct_yaml(
            architectures=[{"build-on": [self.deb_arch], "run-on": [self.deb_arch]}]
        )

        self.run_snapcraft("prime")

        with open(os.path.join("prime", "meta", "snap.yaml")) as f:
            y = yaml_utils.load(f)

        self.assertThat(y["architectures"], Equals([self.deb_arch]))

    def test_default_architecture(self):
        self.construct_yaml(architectures=[{"build-on": ["other-arch"]}])

        self.run_snapcraft("prime")

        with open(os.path.join("prime", "meta", "snap.yaml")) as f:
            y = yaml_utils.load(f)

        self.assertThat(y["architectures"], Equals([self.deb_arch]))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest.mock

from testtools.matchers import Equals

import snapcraft
from snapcraft.internal import errors
from tests import unit


class TestBasePlugin(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.project_options = snapcraft.ProjectOptions()

    def test_cross_compilation_raises(self):
        options = unit.MockOptions(disable_parallel=True)
        plugin = snapcraft.BasePlugin("test_plugin", options, self.project_options)
        self.assertRaises(
            errors.CrossCompilationNotSupported, plugin.enable_cross_compilation
        )

    def test_parallel_build_count_returns_1_when_disabled(self):
        options = unit.MockOptions(disable_parallel=True)
        plugin = snapcraft.BasePlugin("test_plugin", options, self.project_options)
        self.assertThat(plugin.parallel_build_count, Equals(1))

    def test_parallel_build_count_returns_build_count_from_project(self):
        options = unit.MockOptions(disable_parallel=False)
        plugin = snapcraft.BasePlugin("test_plugin", options, self.project_options)
        unittest.mock.patch.object(self.project_options, "parallel_build_count", 2)
        self.assertThat(plugin.parallel_build_count, Equals(2))

    @unittest.mock.patch("snapcraft.internal.common.run")
    def test_run_without_specifying_cwd(self, mock_run):
        plugin = snapcraft.BasePlugin("test/part", options=None)
        plugin.run(["ls"])

        mock_run.assert_called_once_with(["ls"], cwd=plugin.builddir)

    @unittest.mock.patch("snapcraft.internal.common.run")
    def test_run_specifying_a_cwd(self, mock_run):
        plugin = snapcraft.BasePlugin("test/part", options=None)
        plugin.run(["ls"], cwd=plugin.sourcedir)

        mock_run.assert_called_once_with(["ls"], cwd=plugin.sourcedir)

    @unittest.mock.patch("snapcraft.internal.common.run_output")
    def test_run_output_without_specifying_cwd(self, mock_run):
        plugin = snapcraft.BasePlugin("test/part", options=None)
        plugin.run_output(["ls"])

        mock_run.assert_called_once_with(["ls"], cwd=plugin.builddir)

    @unittest.mock.patch("snapcraft.internal.common.run_output")
    def test_run_output_specifying_a_cwd(self, mock_run):
        plugin = snapcraft.BasePlugin("test/part", options=None)
        plugin.run_output(["ls"], cwd=plugin.sourcedir)

        mock_run.assert_called_once_with(["ls"], cwd=plugin.sourcedir)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
from textwrap import dedent

import fixtures
import testtools
from testtools.matchers import EndsWith, Equals, FileContains, FileExists, Not

from tests import integration, skip


class SnapTestCase(integration.TestCase):
    @skip.skip_unless_codename("xenial", "test designed for xenial")
    def test_snap(self):
        self.copy_project_to_cwd("assemble")
        self.run_snapcraft("snap")

        snap_file_path = "assemble_1.0_{}.snap".format(self.deb_arch)
        self.assertThat(snap_file_path, FileExists())

        binary1_wrapper_path = os.path.join(
            self.prime_dir, "command-assemble-bin.wrapper"
        )
        with open("binary1.after", "r") as file_:
            expected_binary1_wrapper = file_.read()
        self.assertThat(binary1_wrapper_path, FileContains(expected_binary1_wrapper))

        self.useFixture(
            fixtures.EnvironmentVariable(
                "SNAP", os.path.join(os.getcwd(), self.prime_dir)
            )
        )
        binary_scenarios = (
            ("command-assemble-service.wrapper", "service-start\n"),
            ("stop-command-assemble-service.wrapper", "service-stop\n"),
            ("command-assemble-bin.wrapper", "binary1\n"),
            ("command-binary2.wrapper", "binary2\n"),
        )
        for binary, expected_output in binary_scenarios:
            output = subprocess.check_output(
                os.path.join(self.prime_dir, binary), universal_newlines=True
            )
            self.assertThat(output, Equals(expected_output))

        with testtools.ExpectedException(subprocess.CalledProcessError):
            subprocess.check_output(
                os.path.join(self.prime_dir, "bin", "not-wrapped"),
                stderr=subprocess.STDOUT,
            )

        self.assertThat(
            os.path.join(self.prime_dir, "bin", "not-wrapped.wrapper"),
            Not(FileExists()),
        )

        self.assertThat(
            os.path.join(
                self.prime_dir, "bin", "command-binary-wrapper-none.wrapper.wrapper"
            ),
            Not(FileExists()),
        )

        # LP: #1750658
        self.assertThat(
            os.path.join(self.prime_dir, "meta", "snap.yaml"),
            FileContains(
                dedent(
                    """\
                name: assemble
                version: '1.0'
                summary: one line summary
                description: a longer description
                architectures:
                - {}
                confinement: strict
                grade: devel
                apps:
                  assemble-bin:
                    command: command-assemble-bin.wrapper
                  assemble-service:
                    command: command-assemble-service.wrapper
                    daemon: simple
                    stop-command: stop-command-assemble-service.wrapper
                  binary-wrapper-none:
                    command: subdir/binary3
                  binary2:
                    command: command-binary2.wrapper
            """
                ).format(self.deb_arch)
            ),
        )

    def test_snap_default(self):
        self.copy_project_to_cwd("assemble")
        self.run_snapcraft([])
        snap_file_path = "assemble_1.0_{}.snap".format(self.deb_arch)
        self.assertThat(snap_file_path, FileExists())

    def test_snap_directory(self):
        self.copy_project_to_cwd("assemble")
        self.run_snapcraft("snap")

        snap_file_path = "assemble_1.0_{}.snap".format(self.deb_arch)
        os.remove(snap_file_path)

        # Verify that Snapcraft can snap its own snap directory (this will make
        # sure `snapcraft snap` and `snapcraft snap <directory>` are always in
        # sync).
        self.run_snapcraft(["snap", "prime"])
        self.assertThat(snap_file_path, FileExists())

    def test_pack_directory(self):
        self.copy_project_to_cwd("assemble")
        self.run_snapcraft("snap")

        snap_file_path = "assemble_1.0_{}.snap".format(self.deb_arch)
        os.remove(snap_file_path)

        # Verify that Snapcraft can snap its own snap directory (this will make
        # sure `snapcraft snap` and `snapcraft pack <directory>` are always
        # in sync).
        self.run_snapcraft(["pack", "prime"])
        self.assertThat(snap_file_path, FileExists())

    def test_snap_long_output_option(self):
        self.run_snapcraft(["snap", "--output", "mysnap.snap"], "assemble")
        self.assertThat("mysnap.snap", FileExists())

    def test_snap_short_output_option(self):
        self.run_snapcraft(["snap", "-o", "mysnap.snap"], "assemble")
        self.assertThat("mysnap.snap", FileExists())

    def test_error_with_inexistent_build_package(self):
        self.copy_project_to_cwd("assemble")
        with open("snapcraft.yaml", "a") as yaml_file:
            yaml_file.write("build-packages:\n  - inexistent-package\n")

        exception = self.assertRaises(
            subprocess.CalledProcessError, self.run_snapcraft, "snap"
        )
        expected = (
            "Could not find a required package in 'build-packages': "
            "inexistent-package\n"
        )
        self.assertThat(exception.output, EndsWith(expected))

    @skip.skip_unless_codename("xenial", "test designed for xenial")
    def test_snap_with_exposed_files(self):
        self.copy_project_to_cwd("nil-plugin-pkgfilter")
        self.run_snapcraft("stage")
        self.assertThat(
            os.path.join(self.stage_dir, "usr", "bin", "nmcli"), FileExists()
        )

        self.run_snapcraft("snap")
        self.assertThat(
            os.path.join(self.prime_dir, "usr", "bin", "nmcli"), FileExists()
        )
        self.assertThat(
            os.path.join(self.prime_dir, "usr", "bin", "nmtui"), Not(FileExists())
        )

    def test_snap_from_snapcraft_init(self):
        self.assertThat("snapcraft.yaml", Not(FileExists()))
        self.run_snapcraft("init")
        self.assertThat(os.path.join("snap", "snapcraft.yaml"), FileExists())

        self.run_snapcraft("snap")

    def test_snap_with_arch(self):
        if self.deb_arch == "armhf":
            self.skipTest("For now, we just support crosscompile from amd64")
        self.run_snapcraft("init")

        self.run_snapcraft(["snap", "--target-arch=i386"])
        self.assertThat("my-snap-name_0.1_i386.snap", FileExists())

    def test_arch_with_snap(self):
        if self.deb_arch == "armhf":
            self.skipTest("For now, we just support crosscompile from amd64")
        self.run_snapcraft("init")

        self.run_snapcraft(["--target-arch=i386", "snap"])
        self.assertThat("my-snap-name_0.1_i386.snap", FileExists())

    def test_implicit_command_with_arch(self):
        if self.deb_arch == "armhf":
            self.skipTest("For now, we just support crosscompile from amd64")
        self.run_snapcraft("init")

        self.run_snapcraft("--target-arch=i386")
        self.assertThat("my-snap-name_0.1_i386.snap", FileExists())

    def test_error_on_bad_yaml(self):
        error = self.assertRaises(
            subprocess.CalledProcessError, self.run_snapcraft, "stage", "bad-yaml"
        )
        self.assertIn(
            "Issues while validating snapcraft.yaml: found character that cannot start "
            "any token on line 13, column 1",
            str(error.output),
        )

    def test_yaml_merge_tag(self):
        self.copy_project_to_cwd("yaml-merge-tag")
        self.run_snapcraft("stage")
        self.assertThat(os.path.join(self.stage_dir, "test.txt"), FileExists())



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib.util
import logging
import os
from importlib.abc import Loader
from typing import Dict  # noqa: F401
from unittest.mock import patch

from snapcraft.extractors import _errors

from ._metadata import ExtractedMetadata

logger = logging.getLogger(__name__)


def extract(relpath: str, *, workdir: str) -> ExtractedMetadata:
    if os.path.basename(relpath) != "setup.py":
        raise _errors.UnhandledFileError(relpath, "setup.py")

    spec = importlib.util.spec_from_file_location(
        "setuppy", os.path.join(workdir, relpath)
    )
    setuppy = importlib.util.module_from_spec(spec)

    params = dict()  # type: Dict[str, str]

    def _fake_setup(*args, **kwargs):
        nonlocal params
        params = kwargs

    with patch("setuptools.setup") as setuptools_mock:
        with patch("distutils.core.setup") as distutils_mock:
            setuptools_mock.side_effect = _fake_setup
            distutils_mock.side_effect = _fake_setup

            # Should never happen, but ensure spec.loader is set.
            loader = spec.loader
            if loader is None or not isinstance(loader, Loader):
                raise RuntimeError("Invalid spec loader")

            # This would really fail during the use of the plugin
            # but let's be cautious and add the proper guards.
            try:
                loader.exec_module(setuppy)
            except SystemExit:
                raise _errors.SetupPyFileParseError(path=relpath)
            except ImportError as e:
                raise _errors.SetupPyImportError(path=relpath, error=str(e)) from e

    version = params.get("version")
    description = params.get("description")

    return ExtractedMetadata(version=version, description=description)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import re

from snapcraft.internal import deprecations
from snapcraft.internal.errors import SnapcraftEnvironmentError
from snapcraft.project import Project

logger = logging.getLogger(__name__)


_EXPECTED_SNAP_DIR_PATTERNS = {
    re.compile(r"^snapcraft.yaml$"),
    re.compile(r"^.snapcraft([/|\\]state)?$"),
    re.compile(r"^hooks([/|\\].*)?$"),
    re.compile(r"^keys([/|\\].*\.(asc))?$"),
    re.compile(r"^local([/|\\].*)?$"),
    re.compile(r"^plugins([/|\\].*)?$"),
    re.compile(r"^gui([/|\\].*\.(png|svg|desktop))?$"),
}


def conduct_project_sanity_check(project: Project, **kwargs) -> None:
    """Sanity check the project itself before continuing.

    The checks done here are meant to be light, and not rely on the build environment.
    """
    # The snapcraft.yaml should be valid even without extensions applied
    # This here check is mostly for backwards compatibility with the
    # rest of the code base.
    if project.info is not None:
        project.info.validate_raw_snapcraft()

    if project._get_build_base() == "core":
        deprecations.handle_deprecation_notice("dn13")

    snap_dir_path = os.path.join(project._get_snapcraft_assets_dir())
    if os.path.isdir(snap_dir_path):
        # TODO: move this check to the ProjectInfo class.
        _check_snap_dir(snap_dir_path)

    if (
        project._get_build_base() in ["core20"]
        and kwargs.get("target_arch") is not None
        and not os.getenv("SNAPCRAFT_ENABLE_EXPERIMENTAL_TARGET_ARCH")
    ):
        raise SnapcraftEnvironmentError(
            "*EXPERIMENTAL* '--target-arch' configured, but not enabled. "
            "Enable with '--enable-experimental-target-arch' flag."
        )

    # Icon should refer to project file, verify it exists.
    icon = project.info.get_raw_snapcraft().get("icon")
    if icon and not os.path.exists(icon):
        raise SnapcraftEnvironmentError(f"Specified icon {icon!r} does not exist.")


def _check_snap_dir(snap_dir_path: str) -> None:
    unexpected_paths = set()
    for root, directories, files in os.walk(snap_dir_path):
        for entry in directories + files:
            path = os.path.relpath(os.path.join(root, entry), snap_dir_path)
            if not _snap_dir_path_expected(path):
                unexpected_paths.add(path)

    if unexpected_paths:
        snap_dir_relpath = os.path.relpath(snap_dir_path, os.getcwd())
        logger.warning(
            "The {snap_dir!r} directory is meant specifically for snapcraft, but it contains "
            "the following non-snapcraft-related paths, which is unsupported and will "
            "cause unexpected behavior:"
            "\n- {unexpected_files}\n\n"
            "If you must store these files within the {snap_dir!r} directory, move them to "
            "{snap_dir_local!r}, which is ignored by snapcraft.".format(
                snap_dir=snap_dir_relpath,
                snap_dir_local=os.path.join(snap_dir_relpath, "local"),
                unexpected_files="\n- ".join(sorted(unexpected_paths)),
            )
        )


def _snap_dir_path_expected(path: str) -> bool:
    for pattern in _EXPECTED_SNAP_DIR_PATTERNS:
        if pattern.match(path):
            return True
    return False



#!/usr/bin/env python3
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import sys

from setuptools import find_namespace_packages, setup


def recursive_data_files(directory, install_directory):
    data_files = []
    for root, directories, file_names in os.walk(directory):
        file_paths = [os.path.join(root, file_name) for file_name in file_names]
        data_files.append((os.path.join(install_directory, root), file_paths))
    return data_files


def determine_version():
    # Examples (git describe -> python package version):
    # 4.1.1-0-gad012482d -> 4.1.1
    # 4.1.1-16-g2d8943dbc -> 4.1.1.post16+g2d8943dbc
    #
    # For shallow clones or repositories missing tags:
    # 0ae7c04

    desc = (
        subprocess.run(
            ["git", "describe", "--always", "--long"],
            check=True,
            stdout=subprocess.PIPE,
        )
        .stdout.decode()
        .strip()
    )

    split_desc = desc.split("-")
    assert (
        len(split_desc) == 3
    ), f"Failed to parse Snapcraft git version description {desc!r}. Confirm that git repository is present and has the required tags/history."

    version = split_desc[0]
    distance = split_desc[1]
    commit = split_desc[2]

    if distance == "0":
        return version

    return f"{version}.post{distance}+git{commit[1:]}"


# Common distribution data
name = "snapcraft"
description = "Publish your app for Linux users for desktop, cloud, and IoT."
author_email = "snapcraft@lists.snapcraft.io"
url = "https://github.com/snapcore/snapcraft"
license = "GPL v3"
classifiers = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: Developers",
    "Intended Audience :: System Administrators",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Natural Language :: English",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Topic :: Software Development :: Build Tools",
    "Topic :: System :: Software Distribution",
]

# snapcraftctl is not in console_scripts because we need a clean environment.
# Only include it for Linux.
if sys.platform == "linux":
    scripts = ["bin/snapcraftctl"]
else:
    scripts = []

dev_requires = [
    "codespell",
    "coverage",
    "flake8==3.7.9",
    "pyflakes==2.1.1",
    "fixtures",
    "isort",
    "mccabe",
    "mypy==0.770",
    "testscenarios",
    "pexpect",
    "pip",
    "pycodestyle==2.5.0",
    "pyftpdlib",
    "pyramid",
    "pytest",
    "pytest-cov",
    "pytest-subprocess",
]

if sys.platform == "win32":
    dev_requires.append("pyinstaller")

install_requires = [
    "attrs",
    "click",
    "cryptography==3.4",
    "gnupg",
    "jsonschema==2.5.1",
    "launchpadlib",
    "lazr.restfulclient",
    "lxml",
    "macaroonbakery",
    "mypy-extensions",
    "progressbar",
    "pyelftools",
    "pymacaroons",
    "pysha3",
    "pyxdg",
    "pyyaml==5.3",
    "raven",
    "requests-toolbelt",
    "requests-unixsocket",
    "requests",
    "simplejson",
    "tabulate",
    "tinydb",
    "typing-extensions",
]

try:
    os_release = open("/etc/os-release").read()
    ubuntu = "ID=ubuntu" in os_release
except FileNotFoundError:
    ubuntu = False

if sys.platform == "linux":
    install_requires += [
        "pylxd",
    ]

if ubuntu:
    install_requires += [
        "catkin-pkg",
        "python-apt",
        "python-debian",
    ]

extras_requires = {
    "dev": dev_requires,
}

setup(
    name=name,
    version=determine_version(),
    description=description,
    author_email=author_email,
    url=url,
    packages=find_namespace_packages(),
    license=license,
    classifiers=classifiers,
    scripts=scripts,
    entry_points=dict(console_scripts=["snapcraft = snapcraft.cli.__main__:run"]),
    data_files=(
        recursive_data_files("schema", "share/snapcraft")
        + recursive_data_files("keyrings", "share/snapcraft")
        + recursive_data_files("extensions", "share/snapcraft")
    ),
    install_requires=install_requires,
    extras_require=extras_requires,
    test_suite="tests.unit",
)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import contextlib
import os
from textwrap import dedent

from snapcraft.internal import errors

_TEMPLATE_YAML = dedent(
    """\
    name: my-snap-name # you probably want to 'snapcraft register <name>'
    base: core18 # the base snap is the execution environment for this snap
    version: '0.1' # just for humans, typically '1.2+git' or '1.3.2'
    summary: Single-line elevator pitch for your amazing snap # 79 char long summary
    description: |
      This is my-snap's description. You have a paragraph or two to tell the
      most important story about your snap. Keep it under 100 words though,
      we live in tweetspace and your description wants to look good in the snap
      store.

    grade: devel # must be 'stable' to release into candidate/stable channels
    confinement: devmode # use 'strict' once you have the right plugs and slots

    parts:
      my-part:
        # See 'snapcraft plugins'
        plugin: nil
    """
)  # noqa, lines too long.


def init():
    """Initialize a snapcraft project."""
    snapcraft_yaml_path = os.path.join("snap", "snapcraft.yaml")

    if os.path.exists(snapcraft_yaml_path):
        raise errors.SnapcraftEnvironmentError(
            "{} already exists!".format(snapcraft_yaml_path)
        )
    elif os.path.exists("snapcraft.yaml"):
        raise errors.SnapcraftEnvironmentError("snapcraft.yaml already exists!")
    elif os.path.exists(".snapcraft.yaml"):
        raise errors.SnapcraftEnvironmentError(".snapcraft.yaml already exists!")
    text = _TEMPLATE_YAML
    with contextlib.suppress(FileExistsError):
        os.mkdir(os.path.dirname(snapcraft_yaml_path))
    with open(snapcraft_yaml_path, mode="w") as f:
        f.write(text)

    return snapcraft_yaml_path



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pathlib
import re
import shutil
import subprocess
from unittest import mock

import pytest
import testtools
from testtools.matchers import Equals

from snapcraft import file_utils
from snapcraft.internal import common, errors
from tests import fixture_setup, unit


class TestReplaceInFile:

    scenarios = [
        (
            "2to3",
            {
                "file_path": os.path.join("bin", "2to3"),
                "contents": "#!/foo/bar/baz/python",
                "expected": "#!/usr/bin/env python",
            },
        ),
        (
            "snapcraft",
            {
                "file_path": os.path.join("bin", "snapcraft"),
                "contents": "#!/foo/baz/python",
                "expected": "#!/usr/bin/env python",
            },
        ),
        (
            "foo",
            {
                "file_path": os.path.join("bin", "foo"),
                "contents": "foo",
                "expected": "foo",
            },
        ),
    ]

    def test_replace_in_file(self, tmp_work_path, file_path, contents, expected):
        (tmp_work_path / "bin").mkdir()

        with open(file_path, "w") as f:
            f.write(contents)

        file_utils.replace_in_file(
            "bin", re.compile(r""), re.compile(r"#!.*python"), r"#!/usr/bin/env python"
        )

        with open(file_path, "r") as f:
            assert f.read() == expected


class TestLinkOrCopyTree(unit.TestCase):
    def setUp(self):
        super().setUp()

        os.makedirs("foo/bar/baz")
        open("1", "w").close()
        open(os.path.join("foo", "2"), "w").close()
        open(os.path.join("foo", "bar", "3"), "w").close()
        open(os.path.join("foo", "bar", "baz", "4"), "w").close()

    def test_link_file_to_file_raises(self):
        raised = self.assertRaises(
            errors.SnapcraftEnvironmentError, file_utils.link_or_copy_tree, "1", "qux"
        )

        self.assertThat(str(raised), Equals("'1' is not a directory"))

    def test_link_file_into_directory(self):
        os.mkdir("qux")
        raised = self.assertRaises(
            errors.SnapcraftEnvironmentError, file_utils.link_or_copy_tree, "1", "qux"
        )

        self.assertThat(str(raised), Equals("'1' is not a directory"))

    def test_link_directory_to_directory(self):
        file_utils.link_or_copy_tree("foo", "qux")
        self.assertTrue(os.path.isfile(os.path.join("qux", "2")))
        self.assertTrue(os.path.isfile(os.path.join("qux", "bar", "3")))
        self.assertTrue(os.path.isfile(os.path.join("qux", "bar", "baz", "4")))

    def test_link_directory_overwrite_file_raises(self):
        open("qux", "w").close()
        raised = self.assertRaises(
            errors.SnapcraftEnvironmentError, file_utils.link_or_copy_tree, "foo", "qux"
        )

        self.assertThat(
            str(raised),
            Equals("Cannot overwrite non-directory 'qux' with directory 'foo'"),
        )

    def test_link_subtree(self):
        file_utils.link_or_copy_tree("foo/bar", "qux")
        self.assertTrue(os.path.isfile(os.path.join("qux", "3")))
        self.assertTrue(os.path.isfile(os.path.join("qux", "baz", "4")))

    def test_link_symlink_to_file(self):
        # Create a symlink to a file
        os.symlink("2", os.path.join("foo", "2-link"))
        file_utils.link_or_copy_tree("foo", "qux")

        # Verify that the symlink remains a symlink
        self.assertThat(os.path.join("qux", "2-link"), unit.LinkExists("2"))

    def test_link_symlink_to_dir(self):
        os.symlink("bar", os.path.join("foo", "bar-link"))
        file_utils.link_or_copy_tree("foo", "qux")

        # Verify that the symlink remains a symlink
        self.assertThat(os.path.join("qux", "bar-link"), unit.LinkExists("bar"))


class TestLinkOrCopy(unit.TestCase):
    def setUp(self):
        super().setUp()

        os.makedirs("foo/bar/baz")
        open("1", "w").close()
        open(os.path.join("foo", "2"), "w").close()
        open(os.path.join("foo", "bar", "3"), "w").close()
        open(os.path.join("foo", "bar", "baz", "4"), "w").close()

    def test_link_file_ioerror(self):
        orig_link = os.link

        def link_and_ioerror(a, b, **kwargs):
            orig_link(a, b)
            raise IOError()

        with mock.patch("os.link") as mock_link:
            mock_link.side_effect = link_and_ioerror
            file_utils.link_or_copy("1", "foo/1")

    def test_copy_nested_file(self):
        file_utils.link_or_copy("foo/bar/baz/4", "foo2/bar/baz/4")
        self.assertTrue(os.path.isfile("foo2/bar/baz/4"))


class RequiresCommandSuccessTestCase(unit.TestCase):
    @mock.patch("subprocess.check_call")
    def test_requires_command_works(self, mock_check_call):
        mock_check_call.side_effect = [None]
        file_utils.requires_command_success("foo").__enter__()
        mock_check_call.assert_called_once_with(
            ["foo"], stderr=subprocess.PIPE, stdout=subprocess.PIPE
        )

    @mock.patch("subprocess.check_call")
    def test_requires_command_multipe_args(self, mock_check_call):
        mock_check_call.side_effect = [None]
        file_utils.requires_command_success("foo bar baz").__enter__()
        mock_check_call.assert_called_once_with(
            ["foo", "bar", "baz"], stderr=subprocess.PIPE, stdout=subprocess.PIPE
        )

    @mock.patch("subprocess.check_call")
    def test_requires_command_success_not_found(self, mock_check_call):
        mock_check_call.side_effect = [FileNotFoundError()]

        raised = self.assertRaises(
            errors.RequiredCommandNotFound,
            file_utils.requires_command_success("foo").__enter__,
        )

        self.assertIsInstance(raised, errors.SnapcraftError)
        self.assertThat(str(raised), Equals("'foo' not found."))

    @mock.patch("subprocess.check_call")
    def test_requires_command_success_error(self, mock_check_call):
        mock_check_call.side_effect = [subprocess.CalledProcessError(1, "x")]

        raised = self.assertRaises(
            errors.RequiredCommandFailure,
            file_utils.requires_command_success("foo").__enter__,
        )

        self.assertIsInstance(raised, errors.SnapcraftError)
        self.assertThat(str(raised), Equals("'foo' failed."))

    def test_requires_command_success_broken(self):
        raised = self.assertRaises(
            TypeError, file_utils.requires_command_success(1).__enter__
        )

        self.assertThat(str(raised), Equals("command must be a string."))

    @mock.patch("subprocess.check_call")
    def test_requires_command_success_custom_error(self, mock_check_call):
        mock_check_call.side_effect = [
            FileNotFoundError(),
            subprocess.CalledProcessError(1, "x"),
        ]

        raised = self.assertRaises(
            errors.RequiredCommandNotFound,
            file_utils.requires_command_success(
                "foo", not_found_fmt="uhm? {cmd_list!r} -> {command}"
            ).__enter__,
        )

        self.assertThat(str(raised), Equals("uhm? ['foo'] -> foo"))

        raised = self.assertRaises(
            errors.RequiredCommandFailure,
            file_utils.requires_command_success(
                "foo", failure_fmt="failed {cmd_list!r} -> {command}"
            ).__enter__,
        )

        self.assertThat(str(raised), Equals("failed ['foo'] -> foo"))


class RequiresPathExistsTestCase(unit.TestCase):
    def setUp(self):
        super(RequiresPathExistsTestCase, self).setUp()
        with open("bar", "w") as fd:
            fd.write("test")

    def test_requires_path_exists_works(self):
        file_utils.requires_path_exists("bar").__enter__()

    def test_requires_path_exists_fails(self):
        raised = self.assertRaises(
            errors.RequiredPathDoesNotExist,
            file_utils.requires_path_exists("foo").__enter__,
        )

        self.assertIsInstance(raised, errors.SnapcraftError)
        self.assertThat(str(raised), Equals("Required path does not exist: 'foo'"))

    def test_requires_path_exists_custom_error(self):
        raised = self.assertRaises(
            errors.RequiredPathDoesNotExist,
            file_utils.requires_path_exists(
                "foo", error_fmt="what? {path!r}"
            ).__enter__,
        )

        self.assertThat(str(raised), Equals("what? 'foo'"))


class TestGetLinkerFromFile(unit.TestCase):
    def test_get_linker_version_from_basename(self):
        self.assertThat(
            file_utils.get_linker_version_from_file("ld-2.26.so"), Equals("2.26")
        )

    def test_get_linker_version_from_path(self):
        self.assertThat(
            file_utils.get_linker_version_from_file("/lib/x86/ld-2.23.so"),
            Equals("2.23"),
        )


class TestGetLinkerFromFileErrors(unit.TestCase):
    def test_bad_file_formatlinker_raises_exception(self):
        self.assertRaises(
            errors.SnapcraftEnvironmentError,
            file_utils.get_linker_version_from_file,
            linker_file="lib64/ld-linux-x86-64.so.2",
        )


_BIN_PATHS = [
    os.path.join("usr", "local", "sbin"),
    os.path.join("usr", "local", "bin"),
    os.path.join("usr", "sbin"),
    os.path.join("usr", "bin"),
    os.path.join("sbin"),
    os.path.join("bin"),
]


class TestGetToolPath:

    scenarios = [
        (i, dict(tool_path=pathlib.Path(i) / "tool-command")) for i in _BIN_PATHS
    ]

    def test_get_tool_from_host_path(self, monkeypatch, tool_path, fake_exists):
        abs_tool_path = pathlib.Path("/") / tool_path
        fake_exists.paths = [abs_tool_path]
        monkeypatch.setattr(shutil, "which", lambda x: abs_tool_path.as_posix())

        assert file_utils.get_snap_tool_path("tool-command") == abs_tool_path.as_posix()

    def test_get_tool_from_snapcraft_snap_path(self, in_snap, tool_path, fake_exists):
        abs_tool_path = pathlib.Path("/snap/snapcraft/current") / tool_path
        fake_exists.paths = [abs_tool_path]

        assert file_utils.get_snap_tool_path("tool-command") == abs_tool_path.as_posix()

    def test_get_tool_from_docker_snap_path(
        self, monkeypatch, in_snap, tool_path, fake_exists
    ):
        abs_tool_path = pathlib.Path("/snap/snapcraft/current") / tool_path
        fake_exists.paths = [abs_tool_path]
        monkeypatch.setattr(common, "is_process_container", lambda: True)

        assert file_utils.get_snap_tool_path("tool-command") == abs_tool_path.as_posix()


def test_get_host_tool_finds_command(monkeypatch):
    monkeypatch.setattr(shutil, "which", lambda x: "/usr/bin/foo")

    assert file_utils.get_host_tool_path(command_name="foo", package_name="foo")


def test_get_host_tool_failure(monkeypatch):
    monkeypatch.setattr(shutil, "which", lambda x: None)

    with pytest.raises(errors.SnapcraftHostToolNotFoundError) as error:
        file_utils.get_host_tool_path(command_name="foo", package_name="foo-pkg")

        assert error.command_name == "foo"
        assert error.package_name == "foo-pkg"


class GetToolPathErrorsTest(testtools.TestCase):
    def test_get_snap_tool_path_fails(self):
        self.assertRaises(
            errors.ToolMissingError,
            file_utils.get_snap_tool_path,
            "non-existent-tool-command",
        )

    def test_get_snap_tool_path_in_container_fails_root(self):
        self.useFixture(fixture_setup.FakeSnapcraftIsASnap())

        self.assertRaises(
            errors.ToolMissingError,
            file_utils.get_snap_tool_path,
            "non-existent-tool-command",
        )



import rclpy
from std_msgs.msg import String


class ListenerNode:
    def __init__(self, node_name):
        self._node = rclpy.Node(node_name)
        self._node.create_subscription(String, "chatter", self.callback)

    def callback(self, message):
        print("I heard: {}".format(message.data))
        self._node.destroy_node()
        rclpy.shutdown()

    def run(self):
        rclpy.spin(self._node)


def main():
    rclpy.init()

    node = ListenerNode("listener")
    node.run()


if __name__ == "__main__":
    main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from testtools.matchers import DirExists, Not

from tests import integration


class CleanTestCase(integration.TestCase):
    def test_clean(self):
        self.copy_project_to_cwd("make-hello")
        self.run_snapcraft("snap")

        snap_dirs = (self.parts_dir, self.stage_dir, self.prime_dir)
        for dir_ in snap_dirs:
            self.assertThat(dir_, DirExists())

        self.run_snapcraft("clean")
        for dir_ in snap_dirs:
            self.assertThat(dir_, Not(DirExists()))

    def test_clean_again(self):
        # Clean a second time doesn't fail.
        # Regression test for https://bugs.launchpad.net/snapcraft/+bug/1497371
        self.copy_project_to_cwd("make-hello")
        self.run_snapcraft("snap")
        self.run_snapcraft("clean")
        self.run_snapcraft("clean")

    # Regression test for LP: #1596596
    def test_clean_invalid_yaml(self):
        self.run_snapcraft("clean", "invalid-snap")
        self.assertThat(self.parts_dir, Not(DirExists()))
        self.assertThat(self.stage_dir, Not(DirExists()))
        self.assertThat(self.prime_dir, Not(DirExists()))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from textwrap import dedent

import fixtures
from testtools.matchers import Contains, Equals

from snapcraft import storeapi

from . import FakeStoreCommandsBaseTestCase


class RevisionsCommandTestCase(FakeStoreCommandsBaseTestCase):

    command_name = "list-revisions"

    def test_revisions_without_snap_raises_exception(self):
        result = self.run_command([self.command_name])

        self.assertThat(result.exit_code, Equals(2))
        self.assertThat(result.output, Contains("Usage:"))

    def test_revisions_without_login_must_ask(self):
        self.fake_store_get_releases.mock.side_effect = [
            storeapi.http_clients.errors.InvalidCredentialsError("error"),
            self.releases,
        ]

        result = self.run_command(
            [self.command_name, "snap-test"], input="user@example.com\nsecret\n"
        )
        self.assertThat(
            result.output, Contains("You are required to login before continuing.")
        )

    def test_revisions(self):
        result = self.run_command([self.command_name, "snap-test"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Contains(
                dedent(
                    """\
            Rev.    Uploaded              Arches    Version    Channels
            2       2016-09-27T19:23:40Z  i386      2.0.1      -
            1       2016-09-27T18:38:43Z  amd64     2.0.2      latest/edge*,latest/stable"""
                )
            ),
        )

    def test_alias(self):
        self.command_name = "list-revisions"

        result = self.run_command([self.command_name, "snap-test"])

        self.assertThat(result.exit_code, Equals(0))

    def test_revisions_by_arch(self):
        result = self.run_command([self.command_name, "snap-test", "--arch=amd64"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Contains(
                dedent(
                    """\
            Rev.    Uploaded              Arches    Version    Channels
            1       2016-09-27T18:38:43Z  amd64     2.0.2      latest/edge*,latest/stable"""
                )
            ),
        )


class DeprecatedHistoryCommandTestCase(FakeStoreCommandsBaseTestCase):
    def test_history_with_deprecation_message(self):
        fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(fake_logger)

        result = self.run_command(["history", "snap-test"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            fake_logger.output,
            Contains(
                "DEPRECATED: The 'history' command has been replaced by "
                "'list-revisions'."
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017,2020-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

import snapcraft
from snapcraft import project
from snapcraft.internal import os_release
from snapcraft.internal.errors import OsReleaseNameError, OsReleaseVersionIdError


def _is_ci_env():
    env_prefixes = ["TRAVIS", "AUTOPKGTEST_TMP"]
    matches = []

    for prefix in env_prefixes:
        matches += [var for var in os.environ.keys() if var.startswith(prefix)]

    return len(matches) > 0


def _get_linux_release(release: os_release.OsRelease) -> str:
    try:
        os_name = release.name()
    except OsReleaseNameError:
        os_name = "Unknown"
    try:
        os_version_id = release.version_id()
    except OsReleaseVersionIdError:
        os_version_id = "Unknown Version"

    return f"{os_name}/{os_version_id}"


def get_user_agent(platform: str = sys.platform) -> str:
    arch = project.Project().deb_arch
    testing = "(testing) " if _is_ci_env() else ""

    if platform == "linux":
        os_platform = _get_linux_release(os_release.OsRelease())
    else:
        os_platform = platform.title()

    return f"snapcraft/{snapcraft.__version__} {testing}{os_platform} ({arch})"
